# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
from pathlib import Path
from textwrap import dedent

import pytest
import yaml

from tesseract_core.sdk.api_parse import ValidationError, validate_tesseract_api

LOGGER = logging.getLogger(__name__)


@pytest.fixture
def valid_tesseract_api() -> str:
    return dedent(
        """
        from typing import List
        from pydantic import BaseModel, Field

        class InputSchema(BaseModel):
            a: List[float] = Field(description="Just an argument.")

        class OutputSchema(BaseModel):
            result: List[float] = Field(description="Whatever.")

        def apply(inputs: InputSchema) -> OutputSchema:
            return OutputSchema(result=inputs.a)

        def abstract_eval(abstract_inputs):
            return {"result": abstract_inputs["a"]}

        # This isn't runtime-valid, but should pass static checks
        def jacobian(inputs: InputSchema, jac_inputs: set[str], jac_outputs: set[str]):
            pass
        """
    )


@pytest.fixture
def valid_tesseract_config() -> str:
    return dedent(
        """
        name: foo
        version: "0+unknown"

        build_config:
            package_data:
              - ["path/to/source", "path/to/destination"]

            custom_build_steps:
              - RUN echo "Hello, World!"
        """
    )


def _write_tesseract_api_to_file(tesseract_api: str, path: Path):
    apifile = path / "tesseract_api.py"
    with open(apifile, "w") as file:
        file.write(tesseract_api)


def _write_tesseract_config_to_file(tesseract_config: str, path: Path):
    configfile = path / "tesseract_config.yaml"
    with open(configfile, "w") as file:
        file.write(tesseract_config)


def test_valid_input_passes_checks(
    tmp_path, valid_tesseract_api, valid_tesseract_config
):
    _write_tesseract_api_to_file(valid_tesseract_api, tmp_path)
    _write_tesseract_config_to_file(valid_tesseract_config, tmp_path)
    validate_tesseract_api(tmp_path)


def test_invalid_config_error(tmp_path, valid_tesseract_api, valid_tesseract_config):
    _write_tesseract_api_to_file(valid_tesseract_api, tmp_path)

    invalid_config = yaml.safe_load(valid_tesseract_config)
    invalid_config["version"] = 1
    _write_tesseract_config_to_file(yaml.dump(invalid_config), tmp_path)

    with pytest.raises(ValidationError, match="should be a valid string"):
        validate_tesseract_api(tmp_path)

    invalid_config = yaml.safe_load(valid_tesseract_config)
    invalid_config["build_config"]["custom_build_steps"] = [1]
    _write_tesseract_config_to_file(yaml.dump(invalid_config), tmp_path)

    with pytest.raises(ValidationError, match="should be a valid string"):
        validate_tesseract_api(tmp_path)

    invalid_config = yaml.safe_load(valid_tesseract_config)
    invalid_config["build_config"]["package_data"] = [["/etc/shadow", "/tmp"]]
    _write_tesseract_config_to_file(yaml.dump(invalid_config), tmp_path)

    with pytest.raises(ValidationError, match="must be a relative path"):
        validate_tesseract_api(tmp_path)


def test_api_not_defined_raises_filenotfound():
    path = Path("/non/existent/path")

    with pytest.raises(
        ValidationError,
        match="No file found at",
    ):
        validate_tesseract_api(path)


def test_invalid_syntax(tmp_path, valid_tesseract_config):
    tesseract_api = "!bad syntax!"
    _write_tesseract_api_to_file(tesseract_api, tmp_path)
    _write_tesseract_config_to_file(valid_tesseract_config, tmp_path)

    with pytest.raises(ValidationError, match="Syntax error"):
        validate_tesseract_api(tmp_path)


def test_missing_required_definition_error(
    tmp_path, valid_tesseract_api, valid_tesseract_config
):
    tesseract_api = valid_tesseract_api.replace("apply", "foobar")

    _write_tesseract_api_to_file(tesseract_api, tmp_path)
    _write_tesseract_config_to_file(valid_tesseract_config, tmp_path)

    with pytest.raises(ValidationError, match="apply not defined"):
        validate_tesseract_api(tmp_path)


def test_apply_signature_errors(tmp_path, valid_tesseract_api, valid_tesseract_config):
    _write_tesseract_config_to_file(valid_tesseract_config, tmp_path)

    tesseract_api = valid_tesseract_api.replace("apply(inputs:", "apply(*args, inputs:")
    _write_tesseract_api_to_file(tesseract_api, tmp_path)
    with pytest.raises(ValidationError, match="keyword-only"):
        validate_tesseract_api(tmp_path)

    tesseract_api = valid_tesseract_api.replace("apply(inputs: ", "apply(*, inputs:")
    _write_tesseract_api_to_file(tesseract_api, tmp_path)
    with pytest.raises(ValidationError, match="keyword-only"):
        validate_tesseract_api(tmp_path)

    tesseract_api = valid_tesseract_api.replace(
        "apply(inputs: InputSchema", "apply(inputs: InputSchema, /"
    )
    _write_tesseract_api_to_file(tesseract_api, tmp_path)
    with pytest.raises(ValidationError, match="positional-only"):
        validate_tesseract_api(tmp_path)

    tesseract_api = valid_tesseract_api.replace(
        "(inputs: InputSchema)",
        "(inputs: InputSchema, sneezy)",
    )
    _write_tesseract_api_to_file(tesseract_api, tmp_path)
    with pytest.raises(ValidationError, match="apply must have 1 argument"):
        validate_tesseract_api(tmp_path)


def test_optional_signature_check(
    tmp_path, valid_tesseract_api, valid_tesseract_config
):
    _write_tesseract_config_to_file(valid_tesseract_config, tmp_path)

    tesseract_api = valid_tesseract_api.replace(
        "jacobian(inputs:", "jacobian(foo, inputs:"
    )
    _write_tesseract_api_to_file(tesseract_api, tmp_path)
    with pytest.raises(ValidationError, match="jacobian must have 3 arguments"):
        validate_tesseract_api(tmp_path)


def test_schema_parent_class_is_checked(
    tmp_path, valid_tesseract_api, valid_tesseract_config
):
    for schema in ("InputSchema", "OutputSchema"):
        tesseract_api = valid_tesseract_api.replace(f"{schema}(BaseModel)", schema)
        _write_tesseract_api_to_file(tesseract_api, tmp_path)
        _write_tesseract_config_to_file(valid_tesseract_config, tmp_path)

        with pytest.raises(
            ValidationError, match=f"{schema} must inherit from pydantic.BaseModel"
        ):
            validate_tesseract_api(tmp_path)



from types import ModuleType

import numpy as np
import pytest
from pydantic import BaseModel

from tesseract_core.runtime import Array, Differentiable, Float32
from tesseract_core.runtime.finite_differences import check_gradients
from tesseract_core.runtime.tree_transforms import get_at_path


class DummyModule(ModuleType):
    def __init__(self, *args, correct_gradients: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self.correct_gradients = correct_gradients

    class InputSchema(BaseModel):
        in_data: Differentiable[Array[(None, 3), Float32]]
        in_dict: dict[str, Differentiable[Array[(3, 3, 3), Float32]]]
        in_scalar: Differentiable[Float32]
        in_aux: str

    class OutputSchema(BaseModel):
        out_data: Differentiable[Array[(None, 3), Float32]]
        out_dict: dict[str, Differentiable[Array[(3, 3, 3), Float32]]]
        out_scalar: Differentiable[Float32]
        out_aux: str

    def apply(self, inputs: InputSchema) -> OutputSchema:
        return {
            "out_data": np.zeros_like(inputs.in_data),
            "out_dict": {
                key: np.zeros_like(value) for key, value in inputs.in_dict.items()
            },
            "out_scalar": np.zeros_like(inputs.in_scalar),
            "out_aux": inputs.in_aux,
        }

    def jacobian(
        self,
        inputs: InputSchema,
        jac_inputs: set[str],
        jac_outputs: set[str],
    ):
        outputs = self.apply(inputs)
        if self.correct_gradients:
            make_array = np.zeros
        else:
            make_array = np.ones

        return {
            key_out: {
                key_in: make_array(
                    (
                        *get_at_path(outputs, key_out).shape,
                        *get_at_path(inputs, key_in).shape,
                    )
                )
                for key_in in jac_inputs
            }
            for key_out in jac_outputs
        }

    def jacobian_vector_product(
        self,
        inputs: InputSchema,
        jvp_inputs: set[str],
        jvp_outputs: set[str],
        tangent_vector: Array[(None,), Float32],
    ):
        outputs = self.apply(inputs)
        if self.correct_gradients:
            make_array = np.zeros
        else:
            make_array = np.ones
        return {
            key_out: make_array(get_at_path(outputs, key_out).shape)
            for key_out in jvp_outputs
        }

    def vector_jacobian_product(
        self,
        inputs: InputSchema,
        vjp_inputs: set[str],
        vjp_outputs: set[str],
        cotangent_vector: Array[(None,), Float32],
    ):
        if self.correct_gradients:
            make_array = np.zeros
        else:
            make_array = np.ones
        return {
            key_in: make_array(get_at_path(inputs, key_in).shape)
            for key_in in vjp_inputs
        }


rng = np.random.default_rng(0)
input_data = {
    "in_data": rng.random((10, 3)),
    "in_dict": {"key": rng.random((3, 3, 3))},
    "in_scalar": rng.random(1)[0],
    "in_aux": "auxiliary",
}


@pytest.mark.parametrize("input_paths", [None, ["in_data"], ["in_dict.{key}"]])
@pytest.mark.parametrize("output_paths", [None, ["out_data", "out_dict.{key}"]])
@pytest.mark.parametrize("endpoints", [None, ["jacobian"]])
def test_check_gradients(input_paths, output_paths, endpoints):
    dummy_module_bad = DummyModule("dummy_module", correct_gradients=False)

    result_iter = check_gradients(
        dummy_module_bad,
        {"inputs": input_data},
        base_dir=None,
        input_paths=input_paths,
        output_paths=output_paths,
        endpoints=endpoints,
        max_evals=10,
    )

    run_endpoints = []
    for endpoint, failures, num_evals in result_iter:
        run_endpoints.append(endpoint)

        # everything should fail (all gradients are wrong)
        assert len(failures) == num_evals

        for failure in failures:
            assert not failure.exception

    # Now try again with correct gradients
    dummy_module_good = DummyModule("dummy_module", correct_gradients=True)
    result_iter = check_gradients(
        dummy_module_good,
        {"inputs": input_data},
        base_dir=None,
        input_paths=input_paths,
        output_paths=output_paths,
        endpoints=endpoints,
        max_evals=10,
    )

    for _, failures, _ in result_iter:
        assert not failures

    if endpoints is not None:
        assert run_endpoints == endpoints
    else:
        assert run_endpoints == [
            "jacobian",
            "jacobian_vector_product",
            "vector_jacobian_product",
        ]



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""End to end tests for docker cli wrapper."""

import os
import subprocess
import textwrap
from contextlib import closing
from pathlib import Path

import docker
import pytest
from common import image_exists

from tesseract_core.sdk.docker_client import (
    APIError,
    ContainerError,
    ImageNotFound,
    build_docker_image,
    is_podman,
)


@pytest.fixture(scope="module")
def docker_py_client():
    # Create a Docker client using the docker-py library
    # You may need to set the environment variable $DOCKER_HOST if encountering issues
    with closing(docker.from_env()) as client:
        yield client


@pytest.fixture()
def docker_client_built_image_name(
    docker_client,
    dummy_tesseract_location,
    dummy_docker_file,
):
    """Build the dummy image for the tests."""
    image_name = "docker_client_create_image_test:dummy"

    build_docker_image(dummy_tesseract_location, image_name, dummy_docker_file)
    try:
        yield image_name
    finally:
        try:
            docker_client.images.remove(image_name)
        except ImageNotFound:
            # already removed
            pass


def test_get_image(docker_client, docker_client_built_image_name, docker_py_client):
    """Test image retrieval."""

    def _strip_image_prefix(image_name):
        """Strip the 'sha256:' prefix from the image name if it exists."""
        if image_name.startswith("sha256:"):
            return image_name.split(":")[1]
        return image_name

    # Get the image
    image = docker_client.images.get(docker_client_built_image_name)
    assert image is not None

    docker_py_image = docker_py_client.images.get(docker_client_built_image_name)
    assert docker_py_image is not None

    docker_image_short = docker_client.images.get(image.short_id)
    assert docker_image_short == image

    # Whether images start with sha256: depends on the used Docker implementation (e.g. Podman vs. Docker)
    if docker_py_image.short_id.startswith("sha256:"):
        non_sha_id = docker_py_image.short_id.split(":")[1]
        docker_image_non_sha = docker_client.images.get(non_sha_id)
        assert docker_image_non_sha is not None
        assert docker_image_short.id == docker_image_non_sha.id

    assert _strip_image_prefix(image.id) == _strip_image_prefix(docker_py_image.id)
    assert _strip_image_prefix(image.short_id) == _strip_image_prefix(
        docker_py_image.short_id
    )
    assert image.tags == docker_py_image.tags
    # Check the repr function
    assert (
        str(image)
        == f"Image(id='{image.id}', short_id='{image.short_id}', tags={image.tags}, attrs={image.attrs})"
    )

    # Check that every image listed by docker cli can be found
    # Dangling images are not listed by the client
    image_ids = subprocess.run(
        [
            "docker",
            "images",
            "--filter",
            "dangling=false",
            "-q",
        ],  # List only image IDs
        capture_output=True,
        text=True,
        check=True,
    )
    image_ids = image_ids.stdout.strip().split("\n")
    # Filter list to exclude empty strings.
    image_ids = [image_id for image_id in image_ids if image_id]
    for image_id in image_ids:
        assert image_exists(docker_client, image_id, tesseract_only=False)

    image_list = docker_client.images.list(tesseract_only=False)
    assert image_list is not None
    assert len(image_list) > 0

    docker_py_image_list = docker_py_client.images.list()
    assert docker_py_image_list is not None

    # Check that every image in image_list is also in docker_py_image_list
    docker_py_image_ids = set(
        _strip_image_prefix(img.id) for img in docker_py_image_list if img
    )
    for image in image_list:
        assert _strip_image_prefix(image.id) in docker_py_image_ids


def test_create_image(
    docker_client,
    docker_py_client,
    dummy_tesseract_location,
    dummy_docker_file,
    docker_client_built_image_name,
):
    """Test image building, getting, and removing.

    Validate image existence in both docker_py_client and docker_client to ensure
    handling of names/ids are the same.
    """
    image1, image2 = None, None
    # Create an image
    try:
        image = docker_client.images.get(docker_client_built_image_name)
        assert image is not None

        image_id_obj = docker_client.images.get(image.id)
        image_short_id_obj = docker_client.images.get(image.short_id)
        assert image == image_id_obj
        assert image_short_id_obj == image
        assert image_exists(docker_client, docker_client_built_image_name)
        assert image_exists(docker_py_client, docker_client_built_image_name)

        # Create a second image with no label
        # Check that :latest gets added automatically
        image1_name = "docker_client_create_image_test"
        docker_client.images.buildx(
            dummy_tesseract_location, image1_name, dummy_docker_file
        )
        image1 = docker_client.images.get(image1_name)
        image1_name = image1_name + ":latest"
        assert image1 is not None
        assert image_exists(docker_client, image1_name)
        assert image_exists(docker_py_client, image1_name)

        # Create a third image with prefixed with repo url
        # Check that name gets handled properly
        repo_url = "local_host/foo/bar/"
        image2_name = "docker_client_create_image_url_test"
        docker_client.images.buildx(
            dummy_tesseract_location, repo_url + image2_name, dummy_docker_file
        )
        image2_py = docker_py_client.images.get(repo_url + image2_name)
        assert image2_py is not None

        image2 = docker_client.images.get(repo_url + image2_name)
        assert image2 is not None

    finally:
        # Clean up the images
        if image1:
            docker_client.images.remove(image1_name)
            assert not image_exists(docker_client, image1_name)
            assert not image_exists(docker_py_client, image1_name)

        if image2:
            docker_client.images.remove(repo_url + image2_name)
            assert not image_exists(docker_client, image2_name)
            assert not image_exists(docker_py_client, repo_url + image2_name)


def test_create_container(
    docker_client, docker_py_client, docker_client_built_image_name
):
    """Test container creation, run, logs, and remove."""

    def _strip_image_prefix(image_name):
        """Strip the 'sha256:' prefix from the image name if it exists."""
        if image_name.startswith("sha256:"):
            return image_name.split(":")[1]
        return image_name

    # Create a container
    container, container_py = None, None
    try:
        container = docker_client.containers.run(
            docker_client_built_image_name, ['echo "Hello, Tesseract!"'], detach=True
        )
        assert container is not None
        container_py = docker_py_client.containers.run(
            docker_client_built_image_name, ['echo "Hello, Tesseract!"'], detach=True
        )
        assert container_py is not None
        # Check property fields
        assert container.project_id is None
        assert container.host_port is None
        assert _strip_image_prefix(container_py.image.id) == _strip_image_prefix(
            container.image.id
        )

        container_get = docker_client.containers.get(container.id)
        container_name_get = docker_client.containers.get(container.name)
        container_py_get = docker_py_client.containers.get(container.id)
        container_py_name_get = docker_py_client.containers.get(container.name)
        assert container_get
        assert container_name_get
        assert container_py_get
        assert container_py_name_get

        assert container_name_get == container_get
        # Only compare the key fields of the original container and the ones from get
        # in case container attrs have been updated since creation
        assert container.id == container_get.id
        assert container.name == container_get.name
        # Compare the docker-py container fields vs docker client fields
        assert container_get.id == container_py_get.id
        assert container_get.short_id == container_py_get.short_id
        assert container_get.name == container_py_get.name
        assert container_get.id == container_py_name_get.id

        status = container.wait()
        status_py = container_py.wait()
        assert status["StatusCode"] == 0
        assert status_py["StatusCode"] == 0

        stdout = container.logs(stdout=True, stderr=False)
        stderr = container.logs(stdout=False, stderr=True)
        assert stdout == b"Hello, Tesseract!\n"
        assert stderr == b""

        stdout_py = container_py.logs(stdout=True, stderr=False)
        stderr_py = container_py.logs(stdout=False, stderr=True)
        assert stdout_py == stdout
        assert stderr_py == stderr

    finally:
        # Clean up the container
        if container:
            container.remove(v=True, force=True)
            # Check that the container is removed
            containers = docker_client.containers.list()
            containers_py = docker_py_client.containers.list()
            assert container.id not in containers
            assert container.id not in containers_py

        if container_py:
            container_py.remove(v=True, force=True)
            # Check that the container is removed
            containers = docker_client.containers.list()
            containers_py = docker_py_client.containers.list()
            assert container_py.id not in containers
            assert container_py.id not in containers_py


def test_container_volume_mounts(
    docker_client, docker_cleanup, docker_client_built_image_name, tmp_path
):
    """Test container volume mounts."""
    container1 = None
    # Pytest creates the tmp_path fixture with drwx------ mode, we need others
    # to be able to read and execute the path so the Docker volume is readable
    # from within the container
    tmp_path.chmod(0o0707)

    dest = Path("/foo/")
    bar_file = dest / "hello.txt"
    stdout = docker_client.containers.run(
        docker_client_built_image_name,
        [f"touch {bar_file} && chmod 777 {bar_file} && echo hello"],
        detach=False,
        volumes={tmp_path: {"bind": dest, "mode": "rw"}},
        remove=True,
    )

    assert stdout == b"hello\n"
    # Check file exists in tmp path
    assert (tmp_path / "hello.txt").exists()

    # Check container is removed and there are no running containers associated with the test image
    result = subprocess.run(
        [
            "docker",
            "ps",
            "--filter",
            f"ancestor={docker_client_built_image_name}",
            "-q",
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    assert result.stdout == ""

    # Open tmp_path/hello.txt in write mode
    with open(tmp_path / "hello.txt", "w") as f:
        f.write("hello tesseract\n")

    # Check we can read it in another container
    container1 = docker_client.containers.run(
        docker_client_built_image_name,
        [f"cat {dest}/hello.txt"],
        detach=True,
        volumes={tmp_path: {"bind": dest, "mode": "rw"}},
    )
    docker_cleanup["containers"].append(container1)
    status = container1.wait()
    assert status["StatusCode"] == 0
    stdout = container1.logs(stdout=True, stderr=False)
    assert stdout == b"hello tesseract\n"


def test_compose_up_down(
    docker_client, docker_py_client, tmp_path, docker_client_built_image_name
):
    """Test docker-compose up and down."""
    project_name, project_name1 = None, None
    try:
        compose_file = tmp_path / "docker-compose.yml"
        # Use tail -f /dev/null to keep the container running
        compose_content = textwrap.dedent(f"""
            services:
              test:
                image: {docker_client_built_image_name}
                command: ["echo 'Hello Tesseract' && tail -f /dev/null"]
        """)
        compose_file.write_text(compose_content)
        # Run docker-compose up
        project_name = docker_client.compose.up(
            str(compose_file), "docker_client_compose_test"
        )
        assert project_name == "docker_client_compose_test"

        # Check that project is visible in list
        projects = docker_client.compose.list()
        assert project_name in projects
        assert docker_client.compose.exists(project_name)

        # Get container associated with this project
        containers = projects.get(project_name)
        container = docker_client.containers.get(containers[0])
        assert container is not None
        stdout = container.logs(stdout=True, stderr=False)
        assert stdout == b"Hello Tesseract\n"

        # Get container from docker-py
        container_py = docker_py_client.containers.get(containers[0])
        assert container_py is not None
        stdout_py = container_py.logs(stdout=True, stderr=False)
        assert stdout_py == stdout

        # Create a second project
        project_name1 = docker_client.compose.up(
            str(compose_file), "docker_client_compose_test_1"
        )
        # Check both projects exist
        assert docker_client.compose.exists(project_name)
        assert docker_client.compose.exists(project_name1)

    finally:
        # Remove one project
        if project_name:
            docker_client.compose.down(project_name)
            assert not docker_client.compose.exists(project_name)

        # Remove second project
        if project_name1:
            assert docker_client.compose.exists(project_name1)
            docker_client.compose.down(project_name1)
            assert not docker_client.compose.exists(project_name1)


def test_compose_error(docker_client, tmp_path, docker_client_built_image_name):
    """Test docker-compose error handling."""
    compose_file = tmp_path / "docker-compose.yml"
    # Write a malformed compose file
    compose_content = textwrap.dedent(f"""
        services:
            test:
            image: {docker_client_built_image_name}
    """)
    compose_file.write_text(compose_content)
    with pytest.raises(APIError) as e:
        docker_client.compose.up(str(compose_file), "docker_client_compose_test")
    # Check that the container's logs were printed to stderr
    assert "Failed to start Tesseract container" in str(e.value)


def test_volume_uid_permissions(
    docker_client, docker_client_built_image_name, docker_volume, docker_cleanup
):
    # Set up root only directory
    def run_tesseract_with_volume(cmd: str, user: str = "root:root", volume: str = ""):
        if not volume:
            volume = {docker_volume.name: {"bind": "/bar", "mode": "rw"}}
        return docker_client.containers.run(
            docker_client_built_image_name,
            [cmd],
            remove=True,
            volumes=volume,
            user=user,
        )

    cmd = "mkdir -p /bar && echo hello > /bar/hello.txt && chmod 700 /bar"
    _ = run_tesseract_with_volume(cmd)

    # Try to read the file as UID 1000
    read_cmd = "cat /bar/hello.txt"
    with pytest.raises(ContainerError) as e:
        _ = run_tesseract_with_volume(read_cmd, user="1000:1000")
    assert "Permission denied" in str(e)

    # Try to write to the folder as UID 1000
    write_cmd = "echo hello > /bar/hello_1000.txt && cat /bar/hello_1000.txt"
    with pytest.raises(ContainerError) as e:
        _ = run_tesseract_with_volume(write_cmd, user="1000:1000")
    assert "Permission denied" in str(e)

    # Grant permission to the file
    cmd = "chmod 777 /bar"
    _ = run_tesseract_with_volume(cmd)

    # Try to read the file as UID 1000 again
    stdout = run_tesseract_with_volume(read_cmd, user="1000:1000")
    assert stdout == b"hello\n"

    # Try to write to the folder as UID 1000 again
    stdout = run_tesseract_with_volume(write_cmd, user="1000:1000")
    assert stdout == b"hello\n"

    # Try to copy a file from a volume with permissions 700 to a volume with permission 777
    volume_777 = docker_client.volumes.create(name="docker_client_test_volume_777")
    docker_cleanup["volumes"].append(volume_777)
    volume_args = {
        docker_volume.name: {"bind": "/from", "mode": "rw"},
        volume_777.name: {"bind": "/to", "mode": "rw"},
    }
    cmd = "chmod 700 /from && cp /from/hello.txt /to/hello.txt && cat /to/hello.txt"
    stdout = run_tesseract_with_volume(cmd, volume=volume_args)
    assert stdout == b"hello\n"

    # Try to access files as UID1000
    cmd = "cat /to/hello.txt"
    stdout = run_tesseract_with_volume(cmd, user="1000:1000", volume=volume_args)
    assert stdout == b"hello\n"

    with pytest.raises(ContainerError) as e:
        run_tesseract_with_volume(
            "cat /from/hello_777.txt", user="1000:1000", volume=volume_args
        )
    assert "Permission denied" in str(e)

    cmd = (
        "adduser -D testuser && chmod 777 /from && su - testuser && cat /from/hello.txt"
    )
    stdout = run_tesseract_with_volume(cmd, volume=volume_args)
    assert stdout == b"hello\n"


def test_is_podman():
    """Test is_podman function.

    This assumes that the DOCKER_HOST environment variable is set to a Podman socket
    when running in a Podman environment. This is true on CI, but may deviate on
    local machines.
    """
    real_is_podman = "podman" in os.environ.get("DOCKER_HOST", "")
    assert is_podman() == real_is_podman



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import re
import sys
import types
from collections.abc import Callable, Iterable, Mapping, Sequence
from copy import copy
from typing import (
    Annotated,
    Any,
    ClassVar,
    Literal,
    Optional,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from pydantic import (
    AfterValidator,
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    ValidationInfo,
    create_model,
    field_validator,
)

from .schema_types import (
    Array,
    PydanticArrayAnnotation,
    ShapeDType,
    _is_annotated,
    is_array_annotation,
    is_differentiable,
    safe_issubclass,
)
from .tree_transforms import get_at_path

# Constants to mark sequence and dict indexing in pytree paths
SEQ_INDEX_SENTINEL = object()
DICT_INDEX_SENTINEL = object()

T = TypeVar("T")

# Python has funnily enough two union types now. See https://github.com/python/cpython/issues/105499
# We check against both for compatibility with older versions of Python.
UNION_TYPES = [Union]
if sys.version_info >= (3, 10):
    UNION_TYPES += [types.UnionType]


def _construct_annotated(obj: Any, metadata: Iterable[Any]) -> Any:
    """Construct an Annotated type with the given metadata."""
    out = obj
    # Exploit that repeated Annotated applications are flattened
    for meta in metadata:
        out = Annotated[out, meta]
    return out


def apply_function_to_model_tree(
    Schema: type[BaseModel],
    func: Callable[[type, tuple], type],
    model_prefix: str = "",
    default_model_config: Optional[dict[str, Any]] = None,
) -> type[BaseModel]:
    """Apply a function to all leaves of a Pydantic model, recursing into containers + nested models.

    A leaf is any type annotation that is not a Pydantic model or a container type.

    The given function should take two arguments: the type annotation and the path to it
    in the model tree (as a list of field names). It should return the new type annotation
    for the field, or None to remove the field from the model.

    The notion of path is used to handle nested models and containers. For example, given
    a model like this:

    class MyModel(BaseModel):
        a: list[dict[str, int]]

    The path to the field "a" would be ["a"], and the path to the int type would be
    ["a", SEQ_INDEX_SENTINEL, DICT_INDEX_SENTINEL].
    """
    # Annotation types that should be treated as leaves and not recursed into
    annotated_types_as_leaves = (PydanticArrayAnnotation,)

    if default_model_config is None:
        default_model_config = {}

    seen_models = set()

    def _recurse_over_model_tree(treeobj: Any, path: list[str]) -> Any:
        # Get the origin type of the annotation, e.g. List for List[int]
        origin_type = get_origin(treeobj)
        deprecated_types = ["List", "Dict", "Set", "FrozenSet", "Tuple"]

        if origin_type is not None and origin_type.__name__ in deprecated_types:
            # Raise error if the user tries to use a deprecated type
            raise TypeError(
                f"'{origin_type}' annotation deprecated. Use '{origin_type.lower()}' instead."
            )

        if safe_issubclass(treeobj, BaseModel):
            # Recurse into (nested) pydantic models

            if id(treeobj) in seen_models:
                raise ValueError(
                    f"Recursive model definition detected at path {path}: {treeobj.__name__}. "
                    "Recursive models are not supported as Tesseract schemas."
                )
            seen_models.add(id(treeobj))

            new_fields = {}

            for field_name, field in treeobj.model_fields.items():
                field_type = field.annotation
                if field.metadata:
                    field_type = _construct_annotated(field_type, field.metadata)

                field_path = [*path, field_name]
                new_type = _recurse_over_model_tree(field_type, field_path)
                if new_type is None:
                    continue

                new_field = copy(field)
                # Need to strip off metadata to trigger re-evaluation of the field
                new_field.metadata = []

                new_fields[field_name] = (new_type, new_field)

            # We only forbid encountering the same model twice if it is within the same subtree
            seen_models.remove(id(treeobj))

            # Only override model_config if it is not already present
            # in the pydantic model definition
            model_config = ConfigDict(**default_model_config)
            model_config.update(treeobj.model_config)

            return create_model(
                f"{model_prefix}{treeobj.__name__}",
                **new_fields,
                model_config=(ConfigDict, model_config),
                __base__=treeobj,
            )

        elif _is_annotated(treeobj):
            # Recurse into Annotated types by stripping off right-most (most recent) metadata

            # Using __origin__ and __metadata__ instead of get_args for compatibility with typing_extensions.Annotated
            inner_type = treeobj.__origin__
            *other_meta, current_meta = treeobj.__metadata__

            is_leaf = False
            for t in annotated_types_as_leaves:
                if safe_issubclass(current_meta, t):
                    is_leaf = True
                    break

            if is_leaf:
                # Reached a leaf -> apply func to object as-is, then re-attach add'l metadata
                current_annotation = Annotated[inner_type, current_meta]
                out = func(current_annotation, tuple(path))
                out = _construct_annotated(out, other_meta)
                return out

            # Not a leaf -> recurse into Annotated[inner_type, *other_meta] and re-attach current_meta to the result
            inner_type = _construct_annotated(inner_type, other_meta)
            new_type = _recurse_over_model_tree(inner_type, path)

            if new_type is None:
                return None

            return Annotated[new_type, current_meta]

        elif any(origin_type is t for t in UNION_TYPES):
            # Recurse into Union and Optional (which is a subclass of Union)
            args = get_args(treeobj)
            newargs = [_recurse_over_model_tree(arg, path) for arg in args]
            newargs = [arg for arg in newargs if arg is not None]
            if not newargs:
                return None

            return Union[tuple(newargs)]

        elif safe_issubclass(origin_type, Mapping):
            # Recurse into dict-likes
            args = get_args(treeobj)
            if not args:
                return origin_type

            key_type, value_type = args

            # Append sentinel to path to indicate dict indexing occurred
            new_path = [*path, DICT_INDEX_SENTINEL]

            new_type = _recurse_over_model_tree(value_type, new_path)
            if new_type is None:
                return None

            return origin_type[key_type, new_type]

        elif safe_issubclass(origin_type, Iterable):
            # Recurse into List, Set, Tuple
            value_type = get_args(treeobj)
            if not value_type:
                return origin_type

            # Append sentinel to path to indicate sequence indexing occurred
            new_path = [*path, SEQ_INDEX_SENTINEL]

            new_types = [_recurse_over_model_tree(vt, new_path) for vt in value_type]
            new_types = [nt for nt in new_types if nt is not None]
            if not new_types:
                return None

            return origin_type[tuple(new_types)]

        # Reached a leaf -> apply func
        return func(treeobj, tuple(path))

    return _recurse_over_model_tree(Schema, [])


def _serialize_diffable_arrays(
    obj: dict[tuple, Any],
) -> dict[str, dict[str, Any]]:
    """Convert a dict {path_tuple: array_type} to a dict {path_str: {shape, dtype}}.

    path_str is a string representation of the path, with SEQ_INDEX_SENTINEL and DICT_INDEX_SENTINEL
    replaced by indexing syntax (e.g. `foo.[].{}`).
    """
    serialized = {}
    for pathtuple, value in obj.items():
        shape = value.__metadata__[0].expected_shape
        dtype = value.__metadata__[0].expected_dtype

        # Ensure shape is JSON serializable
        if shape is Ellipsis:
            shape = None
        else:
            shape = tuple(shape)

        # Replace sentinel values with indexing syntax
        str_parts = []
        for part in pathtuple:
            if part is SEQ_INDEX_SENTINEL:
                str_parts.append("[]")
            elif part is DICT_INDEX_SENTINEL:
                str_parts.append("{}")
            else:
                str_parts.append(part)

        serialized[".".join(str_parts)] = {
            "shape": shape,
            "dtype": dtype,
        }

    return serialized


def create_apply_schema(
    InputSchema: type[BaseModel], OutputSchema: type[BaseModel]
) -> tuple[type[BaseModel], type[BaseModel]]:
    """Create the input / output schemas for the /apply endpoint."""
    # We add metadata to the input and output schemas to indicate which fields are differentiable,
    # what their paths are, and which expected shape / dtype they have.
    # This is used internally and by some official clients, but not advertised as part of the public API,
    # so people should not rely on it.
    diffable_input_paths = _get_diffable_arrays(InputSchema)
    diffable_output_paths = _get_diffable_arrays(OutputSchema)

    InputSchema = apply_function_to_model_tree(
        InputSchema,
        lambda x, _: x,
        model_prefix="Apply_",
        default_model_config=dict(extra="forbid"),
    )
    OutputSchema = apply_function_to_model_tree(
        OutputSchema,
        lambda x, _: x,
        model_prefix="Apply_",
        default_model_config=dict(extra="forbid"),
    )

    class ApplyInputSchema(BaseModel):
        inputs: InputSchema = Field(
            ..., description="The input data to apply the Tesseract to."
        )

        differentiable_arrays: ClassVar[dict[str, dict[str, Any]]] = (
            _serialize_diffable_arrays(diffable_input_paths)
        )
        model_config = ConfigDict(
            extra="forbid",
            json_schema_extra={"differentiable_arrays": differentiable_arrays},
        )

    class ApplyOutputSchema(RootModel):
        root: OutputSchema = Field(
            ..., description="The output data from applying the Tesseract."
        )
        differentiable_arrays: ClassVar[dict[str, dict[str, Any]]] = (
            _serialize_diffable_arrays(diffable_output_paths)
        )
        model_config = ConfigDict(
            json_schema_extra={"differentiable_arrays": differentiable_arrays}
        )

    return ApplyInputSchema, ApplyOutputSchema


def create_abstract_eval_schema(
    InputSchema: type[BaseModel], OutputSchema: type[BaseModel]
) -> tuple[type[BaseModel], type[BaseModel]]:
    """Create new Pydantic models that represent abstract versions of the given schemas.

    The returned schemas will have the same structure as the input schemas, but with array
    fields replaced by ShapeDType objects.
    """

    def replace_array_with_shapedtype(obj: T, _: Any) -> Union[T, type[ShapeDType]]:
        if is_array_annotation(obj):
            return ShapeDType.from_array_annotation(obj)
        return obj

    GeneratedInputSchema = apply_function_to_model_tree(
        InputSchema,
        replace_array_with_shapedtype,
        model_prefix="AbstractEval_",
        default_model_config=dict(extra="forbid"),
    )

    GeneratedOutputSchema = apply_function_to_model_tree(
        OutputSchema,
        replace_array_with_shapedtype,
        model_prefix="AbstractEval_",
        default_model_config=dict(extra="forbid"),
    )

    class AbstractInputSchema(BaseModel):
        inputs: GeneratedInputSchema = Field(
            ...,
            description=(
                "The abstract input data to evaluate the Tesseract on. Has the same structure as InputSchema, "
                "but with array fields replaced by ShapeDType."
            ),
        )
        model_config = ConfigDict(extra="forbid")

    class AbstractOutputSchema(RootModel):
        root: GeneratedOutputSchema = Field(
            ...,
            description=(
                "Abstract outputs with the same structure as OutputSchema, but with array fields "
                "replaced by ShapeDType."
            ),
        )

    return AbstractInputSchema, AbstractOutputSchema


def _get_diffable_arrays(schema: type[BaseModel]) -> dict[tuple, Any]:
    """Return a dictionary mapping path patterns of differentiable arrays to their types."""
    diffable_paths = {}

    def add_to_dict_if_diffable(obj: T, path: tuple) -> T:
        if is_differentiable(obj):
            diffable_paths[path] = obj
        return obj

    apply_function_to_model_tree(schema, add_to_dict_if_diffable)
    return diffable_paths


def _path_to_pattern(path: Sequence[Union[str, object]]) -> str:
    """Return a type describing valid paths for all passed paths."""
    # Check if path includes sequence or dict indexing -- in this case, we can't use the
    # path as a literal and need to use a regex pattern instead.
    final_path = []
    is_literal = True
    for part in path:
        if part is SEQ_INDEX_SENTINEL:
            is_literal = False
            part = r"\[-?\d+\]"
        elif part is DICT_INDEX_SENTINEL:
            is_literal = False
            part = r"\{[\w \-]+\}"
        final_path.append(part)

    if is_literal:
        return ".".join(final_path)

    return r"\.".join(final_path)


def _pattern_to_type(pattern: str) -> type:
    """Convert a string pattern (which may be a literal or regex) to a type that can be used for validation."""
    if _is_regex_pattern(pattern):
        return Annotated[str, Field(pattern=pattern)]
    else:
        return Literal[pattern]


def _is_regex_pattern(pattern: str) -> bool:
    # Poor man's regex detection, but it should work for our purposes
    return "\\" in pattern


def input_path_validator(path: str, info: ValidationInfo) -> str:
    """Validate that the given path points to a valid input key."""
    if "[" in path or "{" in path:
        try:
            get_at_path(info.data["inputs"], path)
        except (LookupError, AttributeError) as exc:
            raise ValueError(
                f"Could not find {info.field_name} path {path} in inputs."
            ) from exc
    return path


def create_autodiff_schema(
    InputSchema: type[BaseModel],
    OutputSchema: type[BaseModel],
    ad_flavor: Literal["jacobian", "jvp", "vjp"],
) -> tuple[type[BaseModel], type[BaseModel]]:
    """Generate the input / outputs schemas for AD endpoints like /jacobian, /jacobian_vector_product, etc.

    Returns a tuple (InputSchema, OutputSchema) with the generated schemas.
    """
    diffable_input_paths = _get_diffable_arrays(InputSchema)
    diffable_output_paths = _get_diffable_arrays(OutputSchema)

    if not diffable_input_paths:
        raise RuntimeError("No differentiable inputs found in the input schema")

    if not diffable_output_paths:
        raise RuntimeError("No differentiable outputs found in the output schema")

    diffable_input_patterns = {
        _path_to_pattern(path): obj for path, obj in diffable_input_paths.items()
    }
    diffable_output_patterns = {
        _path_to_pattern(path): obj for path, obj in diffable_output_paths.items()
    }

    diffable_input_type = Union[
        tuple(_pattern_to_type(p) for p in diffable_input_patterns)
    ]
    diffable_output_type = Union[
        tuple(_pattern_to_type(p) for p in diffable_output_patterns)
    ]

    def _find_shape_from_path(path_patterns: dict, concrete_path: str) -> tuple:
        for path_pattern, array_type in path_patterns.items():
            if _is_regex_pattern(path_pattern):
                path_matches = bool(re.match(path_pattern, concrete_path))
            else:
                path_matches = path_pattern == concrete_path

            if path_matches:
                return array_type.__metadata__[0].expected_shape

        raise ValueError(f"Invalid path: {concrete_path}")

    InputSchema = apply_function_to_model_tree(
        InputSchema,
        lambda x, _: x,
        model_prefix=f"{ad_flavor.title()}_",
        default_model_config=dict(extra="forbid"),
    )

    def result_validator(
        result: dict[str, Any], info: ValidationInfo
    ) -> dict[str, Any]:
        """Validate the result of a Jacobian computation.

        Since the structure of the result is already validated in core.py, we only need to check the shapes
        to ensure they match what's expected from the schema.
        """
        if ad_flavor == "jacobian":
            if set(info.context["output_keys"]) != set(result.keys()):
                raise ValueError(
                    "Error when validating output of jacobian:\n"
                    f"Expected keys {info.context['output_keys']} in output; got {set(result.keys())}"
                )
            for output_path, subout in result.items():
                if set(info.context["input_keys"]) != set(subout.keys()):
                    raise ValueError(
                        "Error when validating output of jacobian:\n"
                        "Expected output with structure "
                        f"{{{tuple(info.context['output_keys'])}: {{{tuple(info.context['input_keys'])}: ...}}}}, "
                        f"got {set(subout.keys())} for output key {output_path}."
                    )
                output_shape = _find_shape_from_path(
                    diffable_output_patterns, output_path
                )
                for input_path, arr in subout.items():
                    input_shape = _find_shape_from_path(
                        diffable_input_patterns, input_path
                    )

                    if output_shape is Ellipsis and input_shape is Ellipsis:
                        # Everything goes
                        continue
                    elif output_shape is Ellipsis:
                        expected_shape = (
                            *input_shape,
                            arr.shape[-len(input_shape) :],
                        )
                    elif input_shape is Ellipsis:
                        expected_shape = (
                            *output_shape,
                            *arr.shape[: len(output_shape)],
                        )
                    else:
                        expected_shape = (*output_shape, *input_shape)

                    # We allow both encoded and decoded arrays as schema output
                    if hasattr(arr, "shape"):
                        got_shape = arr.shape
                    else:
                        got_shape = arr["shape"]

                    if len(got_shape) != len(expected_shape):
                        raise ValueError(
                            f"Jacobian result [{output_path}][{input_path}]: "
                            f"Expected shape {expected_shape}, got {got_shape}"
                        )

                    if any(
                        s1 != s2
                        for s1, s2 in zip(got_shape, expected_shape)
                        if s2 is not None
                    ):
                        raise ValueError(
                            f"Jacobian result [{output_path}][{input_path}]: "
                            f"Expected shape {expected_shape}, got {got_shape}"
                        )

        elif ad_flavor == "jvp":
            if set(info.context["output_keys"]) != set(result.keys()):
                raise ValueError(
                    f"Expected keys {info.context['output_keys']} in output; got {set(result.keys())}"
                )
            for output_path, arr in result.items():
                output_shape = _find_shape_from_path(
                    diffable_output_patterns, output_path
                )
                if output_shape is Ellipsis:
                    # Everything goes
                    continue

                expected_shape = output_shape

                if len(arr.shape) != len(expected_shape):
                    raise ValueError(
                        f"JVP result [{output_path}]: Expected shape {expected_shape}, got {arr.shape}"
                    )

                if any(
                    s1 != s2
                    for s1, s2 in zip(arr.shape, expected_shape)
                    if s2 is not None
                ):
                    raise ValueError(
                        f"JVP result [{output_path}]: Expected shape {expected_shape}, got {arr.shape}"
                    )

        elif ad_flavor == "vjp":
            if set(info.context["input_keys"]) != set(result.keys()):
                raise ValueError(
                    f"Expected keys {info.context['input_keys']} in output; got {set(result.keys())}"
                )
            for input_path, arr in result.items():
                input_shape = _find_shape_from_path(diffable_input_patterns, input_path)
                if input_shape is Ellipsis:
                    # Everything goes
                    continue

                expected_shape = input_shape

                if len(arr.shape) != len(expected_shape):
                    raise ValueError(
                        f"VJP result [{input_path}]: Expected shape {expected_shape}, got {arr.shape}"
                    )

                if any(
                    s1 != s2
                    for s1, s2 in zip(arr.shape, expected_shape)
                    if s2 is not None
                ):
                    raise ValueError(
                        f"VJP result [{input_path}]: Expected shape {expected_shape}, got {arr.shape}"
                    )

        return result

    # NOTE (dh): Literal[somevar] is marked as an error by Pylance, but this doesn't throw a runtime error, so I'd
    # prefer to keep this instead of introducing more arcane hacks to create Literal types at runtime.

    if ad_flavor == "jacobian":

        class JacobianInputSchema(BaseModel):
            inputs: InputSchema = Field(
                ..., description="The input data to compute the Jacobian at."
            )
            jac_inputs: set[
                Annotated[diffable_input_type, AfterValidator(input_path_validator)]
            ] = Field(
                ...,
                description="The set of differentiable inputs to compute the Jacobian with respect to.",
                min_length=1,
            )
            jac_outputs: set[diffable_output_type] = Field(
                ...,
                description="The set of differentiable outputs to compute the Jacobian of.",
                min_length=1,
            )

            model_config = ConfigDict(extra="forbid")

        class JacobianOutputSchema(RootModel):
            root: Annotated[
                dict[diffable_output_type, dict[diffable_input_type, Array[..., None]]],
                AfterValidator(result_validator),
            ] = Field(
                ...,
                description=(
                    "Container for the results of Jacobian computations. "
                    "The result represents a nested structure of the Jacobian matrix "
                    "as a mapping with structure ``{jac_outputs: {jac_inputs: array}}``. "
                    "The shape of each array is the concatenation of the shapes of the output and input arrays, "
                    "i.e. ``(*output_array.shape, *input_array.shape)``."
                ),
            )

        InputSchema = JacobianInputSchema
        OutputSchema = JacobianOutputSchema

    elif ad_flavor == "jvp":

        class JVPInputSchema(BaseModel):
            inputs: InputSchema = Field(
                ..., description="The input data to compute the JVP at."
            )
            jvp_inputs: set[
                Annotated[diffable_input_type, AfterValidator(input_path_validator)]
            ] = Field(
                ...,
                description="The set of differentiable inputs to compute the JVP with respect to.",
                min_length=1,
            )
            jvp_outputs: set[diffable_output_type] = Field(
                ...,
                description="The set of differentiable outputs to compute the JVP of.",
                min_length=1,
            )
            tangent_vector: dict[diffable_input_type, Array[..., None]] = Field(
                ...,
                description=(
                    "Tangent vector to multiply the Jacobian with. "
                    "Expected to be a mapping with structure ``{jvp_inputs: array}``. "
                    "The shape of each array is the same as the shape of the corresponding input array."
                ),
            )
            model_config = ConfigDict(extra="forbid")

            @field_validator("tangent_vector", mode="after")
            @classmethod
            def validate_tangent_vector(
                cls, tangent_vector: dict, info: ValidationInfo
            ) -> dict:
                """Raise an exception if the input of an autodiff function does not conform to given input keys."""
                # Cotangent vector needs same keys as output_keys
                try:
                    if set(tangent_vector.keys()) != info.data["jvp_inputs"]:
                        raise ValueError(
                            f"Expected tangent vector with keys conforming to jvp_inputs: {info.data['jvp_inputs']}, "
                            f"got {set(tangent_vector.keys())}."
                        )
                except KeyError as e:
                    raise ValueError(
                        "Unable to validate tangent vector as jvp_inputs either missing or invalid"
                    ) from e
                return tangent_vector

        class JVPOutputSchema(RootModel):
            root: Annotated[
                dict[diffable_output_type, Array[..., None]],
                AfterValidator(result_validator),
            ] = Field(
                ...,
                description=(
                    "Container for the results of Jacobian-vector products. "
                    "The result is a mapping with structure ``{jvp_outputs: array}``. "
                    "The shape of each array is the same as the shape of the corresponding output array."
                ),
            )

        InputSchema = JVPInputSchema
        OutputSchema = JVPOutputSchema

    elif ad_flavor == "vjp":

        class VJPInputSchema(BaseModel):
            inputs: InputSchema = Field(
                ..., description="The input data to compute the VJP at."
            )
            vjp_inputs: set[
                Annotated[diffable_input_type, AfterValidator(input_path_validator)]
            ] = Field(
                ...,
                description="The set of differentiable inputs to compute the VJP with respect to.",
                min_length=1,
            )
            vjp_outputs: set[diffable_output_type] = Field(
                ...,
                description="The set of differentiable outputs to compute the VJP of.",
                min_length=1,
            )
            cotangent_vector: dict[diffable_output_type, Array[..., None]] = Field(
                ...,
                description=(
                    "Cotangent vector to multiply the Jacobian with. "
                    "Expected to be a mapping with structure ``{vjp_outputs: array}``. "
                    "The shape of each array is the same as the shape of the corresponding output array."
                ),
            )
            model_config = ConfigDict(extra="forbid")

            @field_validator("cotangent_vector", mode="after")
            @classmethod
            def validate_cotangent_vector(
                cls, cotangent_vector: dict, info: ValidationInfo
            ) -> dict:
                """Raise an exception if cotangent vector keys are not identical to vjp_outputs."""
                try:
                    if set(cotangent_vector.keys()) != info.data["vjp_outputs"]:
                        raise ValueError(
                            "Expected cotangent vector with keys conforming to vjp_outputs:",
                            f"{info.data['vjp_outputs']}, got {set(cotangent_vector.keys())}.",
                        )
                except KeyError as e:
                    raise ValueError(
                        "Unable to validate cotangent vector as vjp_outputs either missing or invalid"
                    ) from e
                return cotangent_vector

        class VJPOutputSchema(RootModel):
            root: Annotated[
                dict[diffable_input_type, Array[..., None]],
                AfterValidator(result_validator),
            ] = Field(
                ...,
                description=(
                    "Container for the results of vector-Jacobian products. "
                    "The result is a mapping with structure ``{vjp_inputs: array}``. "
                    "The shape of each array is the same as the shape of the corresponding input array."
                ),
            )

        InputSchema = VJPInputSchema
        OutputSchema = VJPOutputSchema

    return InputSchema, OutputSchema



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import traceback
from collections.abc import Iterator, Sequence
from functools import wraps
from pathlib import Path
from types import ModuleType
from typing import (
    Any,
    Callable,
    Literal,
    NamedTuple,
    Optional,
    get_args,
)

import numpy as np
from numpy.typing import ArrayLike
from pydantic import BaseModel
from rich.progress import Progress

from .core import create_endpoints
from .tree_transforms import get_at_path, set_at_path

ADEndpointName = Literal[
    "jacobian", "jacobian_vector_product", "vector_jacobian_product"
]


class GradientCheckResult(NamedTuple):
    """Result of a gradient check (Jacobian row).

    Attributes:
        in_path: The input path of the gradient check.
        out_path: The output path of the gradient check.
        idx: The row index of the gradient check.
        grad_val: The value of the gradient at the given index.
        ref_val: The value of the reference gradient at the given index.
        exception: The exception raised during the gradient check, if any.
    """

    in_path: str
    out_path: str
    idx: tuple[int, ...]
    grad_val: Optional[ArrayLike]
    ref_val: Optional[ArrayLike]
    exception: Optional[str]


def get_input_schema(endpoint_function: Callable) -> type[BaseModel]:
    """Get the input schema of an endpoint function."""
    schema = endpoint_function.__annotations__["payload"]
    if not issubclass(schema, BaseModel):
        raise AssertionError(f"Expected BaseModel, got {schema}")
    return schema


def get_output_schema(endpoint_function: Callable) -> type[BaseModel]:
    """Get the output schema of an endpoint function."""
    schema = endpoint_function.__annotations__["return"]
    if not issubclass(schema, BaseModel):
        raise AssertionError(f"Expected BaseModel, got {schema}")
    return schema


def expand_path_pattern(path_pattern: str, inputs: dict[str, Any]) -> list[str]:
    """Expand a path pattern to a list of all matching paths in the given pytree.

    For example, given the path pattern `a.[].{}`, and the inputs `{"a": [{"b": 1}, {"c": 2}]}`,
    this function would return `["a.[0].{b}", "a.[1].{c}"]`.
    """
    parts = path_pattern.split(".")

    def _handle_part(
        parts: Sequence[str], current_inputs: Any, current_path: list[str]
    ) -> list[str]:
        """Recursively expand each part separately."""
        if not parts:
            return [".".join(current_path)]

        paths = []
        part = parts[0]

        if part == "[]":
            # sequence access
            for i, _ in enumerate(current_inputs):
                subpaths = _handle_part(
                    parts[1:], current_inputs[i], [*current_path, f"[{i}]"]
                )
                paths.extend(subpaths)
        elif part == "{}":
            # dictionary access
            for key in current_inputs:
                subpaths = _handle_part(
                    parts[1:], current_inputs[key], [*current_path, f"{{{key}}}"]
                )
                paths.extend(subpaths)
        else:
            subpaths = _handle_part(
                parts[1:], current_inputs[part], [*current_path, part]
            )
            paths.extend(subpaths)
        return paths

    return _handle_part(parts, inputs, [])


def get_differentiable_paths(
    apply_endpoint_fn: Callable, inputs: dict[str, Any], outputs: dict[str, Any]
) -> tuple[list[str], list[str]]:
    """Get the paths of all differentiable leaves present in the given inputs and outputs."""
    InputSchema = get_input_schema(apply_endpoint_fn)
    OutputSchema = get_output_schema(apply_endpoint_fn)

    diffable_input_paths = InputSchema.differentiable_arrays
    diffable_output_paths = OutputSchema.differentiable_arrays

    ad_inputs = []
    for pattern in diffable_input_paths:
        ad_inputs.extend(expand_path_pattern(pattern, inputs))

    ad_outputs = []
    for pattern in diffable_output_paths:
        ad_outputs.extend(expand_path_pattern(pattern, outputs))

    return ad_inputs, ad_outputs


def _cached_jacobian(fn: Callable) -> Callable:
    """Cache the result of the jacobian computation based on input_path, output_path, and input_idx."""
    cache = {}

    @wraps(fn)
    def _wrapper(*args: Any, **kwargs: Any) -> Any:
        _, _, input_path, output_path, input_idx, *_ = args
        key = (input_path, output_path, tuple(input_idx))
        if key not in cache:
            try:
                cache[key] = fn(*args, **kwargs)
            except Exception as e:
                cache[key] = e
        if isinstance(cache[key], Exception):
            raise cache[key]
        return cache[key]

    _wrapper.clear_cache = cache.clear
    return _wrapper


@_cached_jacobian
def _jacobian_via_apply(
    endpoints_func: dict[str, Callable],
    inputs: dict[str, Any],
    input_path: Sequence[str],
    output_path: Sequence[str],
    input_idx: tuple[int, ...],
    eps: float = 1e-4,
) -> ArrayLike:
    """Compute a Jacobian row using central finite differences."""
    apply_fn = endpoints_func["apply"]
    ApplySchema = get_input_schema(apply_fn)

    def _perturbed_apply(inputs: dict[str, Any], eps: float) -> dict[str, Any]:
        input_val = get_at_path(inputs, input_path).copy()
        if input_idx:
            # array
            input_val[input_idx] += eps
        else:
            # scalar
            input_val += eps
        inputs_plus = set_at_path(inputs, {input_path: input_val})
        return apply_fn(
            ApplySchema.model_validate({"inputs": inputs_plus})
        ).model_dump()

    output_plus = _perturbed_apply(inputs, eps)
    output_minus = _perturbed_apply(inputs, -eps)

    central_diff = (
        get_at_path(output_plus, output_path) - get_at_path(output_minus, output_path)
    ) / (2 * eps)
    return central_diff


@_cached_jacobian
def _jacobian_via_jacobian(
    endpoints_func: dict[str, Callable],
    inputs: dict[str, Any],
    input_path: Sequence[str],
    output_path: Sequence[str],
    input_idx: tuple[int, ...],
) -> ArrayLike:
    """Compute a Jacobian row using the jacobian endpoint."""
    jac_fn = endpoints_func["jacobian"]

    def _jacobian(inputs: dict[str, Any]) -> dict[str, Any]:
        JacSchema = get_input_schema(jac_fn)
        return jac_fn(
            JacSchema.model_validate(
                {
                    "inputs": inputs,
                    "jac_inputs": [input_path],
                    "jac_outputs": [output_path],
                }
            )
        ).model_dump()

    output = _jacobian(inputs)
    output_val = output[output_path][input_path]
    # Jacobian output has shape (*output_shape, *input_shape), where we slice into input_shape
    # while passing through output_shape.
    jac_slice = (slice(None),) * (output_val.ndim - len(input_idx)) + tuple(input_idx)
    return output_val[jac_slice]


@_cached_jacobian
def _jacobian_via_jvp(
    endpoints_func: dict[str, Callable],
    inputs: dict[str, Any],
    input_path: Sequence[str],
    output_path: Sequence[str],
    input_idx: tuple[int, ...],
) -> ArrayLike:
    """Compute a Jacobian row using the jacobian_vector_product endpoint."""
    jvp_fn = endpoints_func["jacobian_vector_product"]
    JvpSchema = get_input_schema(jvp_fn)

    tangent = np.zeros_like(get_at_path(inputs, input_path))
    tangent[input_idx] = 1
    jvp = jvp_fn(
        JvpSchema.model_validate(
            {
                "inputs": inputs,
                "jvp_inputs": [input_path],
                "jvp_outputs": [output_path],
                "tangent_vector": {input_path: tangent},
            }
        )
    ).model_dump()
    return jvp[output_path]


@_cached_jacobian
def _jacobian_via_vjp(
    endpoints_func: dict[str, Callable],
    inputs: dict[str, Any],
    input_path: Sequence[str],
    output_path: Sequence[str],
    input_idx: tuple[int, ...],
) -> ArrayLike:
    """Compute a Jacobian row using the vector_jacobian_product endpoint."""
    apply_fn = endpoints_func["apply"]
    ApplySchema = get_input_schema(apply_fn)
    outputs = apply_fn(ApplySchema.model_validate({"inputs": inputs})).model_dump()

    vjp_fn = endpoints_func["vector_jacobian_product"]
    VjpSchema = get_input_schema(vjp_fn)
    jac_row = np.zeros_like(get_at_path(outputs, output_path))

    for col_idx in np.ndindex(jac_row.shape):
        cotangent = np.zeros_like(jac_row)
        cotangent[col_idx] = 1
        vjp = vjp_fn(
            VjpSchema.model_validate(
                {
                    "inputs": inputs,
                    "vjp_inputs": [input_path],
                    "vjp_outputs": [output_path],
                    "cotangent_vector": {output_path: cotangent},
                }
            )
        ).model_dump()
        jac_row[col_idx] = vjp[input_path][input_idx]

    return jac_row


def _sample_indices(
    inputs: dict[str, Any],
    diff_inputs: list[str],
    diff_outputs: list[str],
    max_evals: int,
    rng: np.random.RandomState,
) -> list[tuple[str, str, tuple[int, ...]]]:
    """Sample combinations of (input_path, output_path, row_idx) to check.

    row_idx are sampled at random, proportional to the size of the input.
    """
    input_shapes = {path: get_at_path(inputs, path).shape for path in diff_inputs}
    total_elements = sum(np.prod(shape) for shape in input_shapes.values())

    idx_per_input = {}
    for path, shape in input_shapes.items():
        if not shape:
            idx_per_input[path] = [()]
            continue
        n_evals = max(1, int(max_evals * np.prod(shape) / total_elements))
        idx_tuple = np.unravel_index(rng.choice(int(np.prod(shape)), n_evals), shape)
        idx_per_input[path] = list(zip(*idx_tuple))

    items_to_check = []
    for in_path in diff_inputs:
        for idx in idx_per_input[in_path]:
            idx = tuple(int(i) for i in idx)
            for out_path in diff_outputs:
                items_to_check.append((in_path, out_path, idx))

    return items_to_check


def check_endpoint_gradients(
    endpoint_functions: dict[str, Callable],
    inputs: dict[str, Any],
    endpoint: str,
    *,
    diff_inputs: list[str],
    diff_outputs: list[str],
    max_evals: int,
    eps: float,
    rtol: float,
    rng: np.random.RandomState,
    show_progress: bool,
) -> tuple[list[GradientCheckResult], int]:
    """Check gradients of an endpoint against a finite difference approximation."""
    failures = []

    if endpoint == "jacobian":
        _jacobian_via_grad = _jacobian_via_jacobian
    elif endpoint == "jacobian_vector_product":
        _jacobian_via_grad = _jacobian_via_jvp
    elif endpoint == "vector_jacobian_product":
        _jacobian_via_grad = _jacobian_via_vjp
    else:
        raise AssertionError(f"Unknown endpoint {endpoint}")

    items_to_check = _sample_indices(inputs, diff_inputs, diff_outputs, max_evals, rng)
    num_evals = 0

    try:
        with Progress(disable=not show_progress) as progress:
            subtask = progress.add_task(
                f"Checking gradients for {endpoint}...", total=len(items_to_check)
            )

            for in_path, out_path, idx in items_to_check:
                num_evals += 1

                failure = None
                try:
                    result_apply = _jacobian_via_apply(
                        endpoint_functions,
                        inputs,
                        in_path,
                        out_path,
                        idx,
                        eps=eps,
                    )
                    result_grad = _jacobian_via_grad(
                        endpoint_functions,
                        inputs,
                        in_path,
                        out_path,
                        idx,
                    )
                except Exception as e:
                    tb = traceback.extract_tb(e.__traceback__)
                    exc_info = f"{type(e).__name__}: '{e}' in file {tb[-1].filename}, line {tb[-1].lineno}"
                    failure = GradientCheckResult(
                        in_path=in_path,
                        out_path=out_path,
                        idx=idx,
                        ref_val=None,
                        grad_val=None,
                        exception=exc_info,
                    )
                else:
                    if not np.allclose(result_apply, result_grad, atol=1e-8, rtol=rtol):
                        failure = GradientCheckResult(
                            in_path=in_path,
                            out_path=out_path,
                            idx=idx,
                            ref_val=result_apply,
                            grad_val=result_grad,
                            exception=None,
                        )

                if failure:
                    failures.append(failure)
                    progress.update(
                        subtask,
                        description=f"Checking gradients for {endpoint}... (failures: {len(failures)})",
                    )

                progress.update(subtask, advance=1)

    except BaseException as e:
        # Sometimes, Pydantic re-raises exceptions as Pydantic<...>Exception so we check the string representation
        is_interrupt = isinstance(e, KeyboardInterrupt) or "KeyboardInterrupt" in str(e)
        if not is_interrupt:
            raise
        print("Interrupted")

    return failures, num_evals


def check_gradients(
    api_module: ModuleType,
    inputs: dict[str, Any],
    *,
    input_paths: Optional[Sequence[str]] = None,
    output_paths: Optional[Sequence[str]] = None,
    base_dir: Optional[Path] = None,
    endpoints: Optional[Sequence[ADEndpointName]] = None,
    max_evals: int = 1000,
    eps: float = 1e-4,
    rtol: float = 0.1,
    seed: Optional[int] = None,
    show_progress: bool = True,
) -> Iterator[tuple[str, list[GradientCheckResult], int]]:
    """Check gradients of endpoints against a finite difference approximation.

    Args:
        api_module: The module containing the Tesseract endpoints.
        inputs: The inputs to apply to evaluate gradients at.
        input_paths: The input paths to check. If not provided, all differentiable paths are checked.
        output_paths: The output paths to check. If not provided, all differentiable paths are checked.
        base_dir: The base directory to resolve relative paths.
        endpoints: The AD endpoints to check. If not provided, all available endpoints are checked.
        max_evals: The target number of ``apply`` evaluations to perform.
        eps: The epsilon to use for finite differences, as a fraction of the maximum absolute value of each input.
        rtol: The relative tolerance to use for comparison.
        seed: The random seed to use for sampling. If not provided, a random seed is used.
        show_progress: Whether to show a progress bar.
    """
    # We apply a global cache to these functions to avoid hashing `inputs` multiple times,
    # so we need to clear the cache before each run.
    _jacobian_via_apply.clear_cache()
    _jacobian_via_jacobian.clear_cache()
    _jacobian_via_jvp.clear_cache()
    _jacobian_via_vjp.clear_cache()

    # Get available endpoints
    endpoint_functions = {func.__name__: func for func in create_endpoints(api_module)}
    available_endpoints = [
        func_name
        for func_name in endpoint_functions
        if func_name in get_args(ADEndpointName)
    ]

    if not available_endpoints:
        raise ValueError(f"No AD endpoints found in {api_module.__name__}")

    if not endpoints:
        endpoints = available_endpoints

    for endpoint in endpoints:
        if endpoint not in available_endpoints:
            raise ValueError(f"Endpoint {endpoint} not found in {api_module.__name__}")

    # Load + dump inputs to ensure they are valid + normalized
    InputSchema = get_input_schema(endpoint_functions["apply"])
    loaded_inputs = InputSchema.model_validate(inputs, context={"base_dir": base_dir})
    inputs = loaded_inputs.inputs.model_dump()
    outputs = endpoint_functions["apply"](loaded_inputs).model_dump()

    # Get differentiable paths
    diff_inputs, diff_outputs = get_differentiable_paths(
        endpoint_functions["apply"],
        inputs,
        outputs,
    )

    if not input_paths:
        input_paths = diff_inputs

    for path in input_paths:
        if path not in diff_inputs:
            raise ValueError(
                f"Input path {path} not found in differentiable paths ({diff_inputs})"
            )

    if not output_paths:
        output_paths = diff_outputs

    for path in output_paths:
        if path not in diff_outputs:
            raise ValueError(
                f"Output path {path} not found in differentiable paths ({diff_outputs})"
            )

    # Check gradients for each endpoint separately
    rng = np.random.RandomState(seed)

    for endpoint in endpoints:
        failures, num_evals = check_endpoint_gradients(
            endpoint_functions,
            inputs,
            endpoint,
            diff_inputs=input_paths,
            diff_outputs=output_paths,
            max_evals=max_evals,
            eps=eps,
            rtol=rtol,
            rng=rng,
            show_progress=show_progress,
        )
        yield endpoint, failures, num_evals



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import pyvista as pv
from pydantic import BaseModel, Field

from tesseract_core.runtime import Array, Float32


class MeshInfo(BaseModel):
    dimensions: list[int]
    n_points: int
    n_cells: int
    bounds: list[float]
    center: list[float]


class InputSchema(BaseModel):
    vertices: Array[(None, 3), Float32] = Field(
        description="3D vector of the mesh vertices."
    )


class OutputSchema(BaseModel):
    pv_mesh: MeshInfo = Field(description="PyVista mesh object info.")


def apply(inputs: InputSchema) -> OutputSchema:
    """Convert input vertices into a PyVista mesh."""
    grid = pv.StructuredGrid(*inputs.vertices)
    info = MeshInfo(
        dimensions=grid.dimensions,
        n_points=grid.n_points,
        n_cells=grid.n_cells,
        bounds=grid.bounds,
        center=grid.center,
    )
    return OutputSchema(pv_mesh=info)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""CLI tests that do not require a running Docker daemon.

(Those go in endtoend_tests/test_endtoend.py.)
"""

import os
import subprocess

import pytest
from typer.testing import CliRunner

from tesseract_core.sdk.cli import app as cli


@pytest.fixture
def cli_runner():
    return CliRunner(mix_stderr=False)


def test_suggestion_on_misspelled_command(cli_runner):
    result = cli_runner.invoke(cli, ["innit"], catch_exceptions=False)
    assert result.exit_code == 2, result.stdout
    assert "No such command 'innit'." in result.stderr
    assert "Did you mean 'init'?" in result.stderr

    result = cli_runner.invoke(cli, ["wellbloodygreatinnit"], catch_exceptions=False)
    assert result.exit_code == 2, result.stdout
    assert "No such command 'wellbloodygreatinnit'." in result.stderr
    assert "Did you mean" not in result.stderr


def test_version(cli_runner):
    from tesseract_core import __version__

    result = cli_runner.invoke(cli, ["--version"])
    assert result.exit_code == 0, result.stdout
    assert __version__ in result.stdout


def test_bad_docker_executable_env_var():
    env = os.environ.copy()
    env.update({"TESSERACT_DOCKER_EXECUTABLE": "not-a-docker"})

    result = subprocess.run(
        ["tesseract", "ps"],
        env=env,
        check=False,
        capture_output=True,
    )
    assert result.returncode == 1
    assert "Executable `not-a-docker` not found" in result.stderr.decode()


@pytest.mark.parametrize(
    "arg_to_override",
    [
        "name",
        "build_config.custom_build_steps",
        "build_config.base_image",
        "build_config.package_data",
    ],
)
def test_config_override(
    arg_to_override, cli_runner, mocker, dummy_tesseract_location, mocked_docker
):
    mocked_build = mocker.patch("tesseract_core.sdk.engine.build_tesseract")

    def _run_with_override(key, value):
        return cli_runner.invoke(
            cli,
            [
                "build",
                str(dummy_tesseract_location),
                "--config-override",
                f"{key}={value}",
                "--generate-only",
            ],
            catch_exceptions=False,
        )

    if arg_to_override == "name":
        argpairs = (
            (
                "my-tesseract",
                {("name",): "my-tesseract"},
            ),
        )
    elif arg_to_override == "build_config.custom_build_steps":
        argpairs = (
            (
                "[RUN foo='bar']",
                {("build_config", "custom_build_steps"): ["RUN foo='bar'"]},
            ),
            (
                '[RUN echo "hello world"]',
                {("build_config", "custom_build_steps"): ['RUN echo "hello world"']},
            ),
        )
    elif arg_to_override == "build_config.base_image":
        argpairs = (
            (
                "ubuntu:latest",
                {("build_config", "base_image"): "ubuntu:latest"},
            ),
        )
    elif arg_to_override == "build_config.package_data":
        argpairs = (
            (
                '["data/file.txt:/app/data/file.txt"]',
                {
                    ("build_config", "package_data"): [
                        "data/file.txt:/app/data/file.txt"
                    ]
                },
            ),
        )
    else:
        raise ValueError(f"Unknown arg_to_override: {arg_to_override}")

    for value, expected in argpairs:
        result = _run_with_override(arg_to_override, value)
        assert result.exit_code == 0, result.stderr
        assert mocked_build.call_args[1]["config_override"] == expected



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Engine to power Tesseract commands."""

import datetime
import linecache
import logging
import optparse
import os
import random
import re
import socket
import string
import tempfile
import threading
import time
from collections.abc import Callable, Sequence
from contextlib import closing
from pathlib import Path
from shutil import copy, copytree, rmtree
from typing import Any

import requests
from jinja2 import Environment, PackageLoader, StrictUndefined
from pip._internal.index.package_finder import PackageFinder
from pip._internal.network.session import PipSession
from pip._internal.req.req_file import (
    RequirementsFileParser,
    get_line_parser,
    handle_line,
)

from .api_parse import TesseractConfig, get_config, validate_tesseract_api
from .docker_client import (
    APIError,
    CLIDockerClient,
    Container,
    ContainerError,
    Image,
    NotFound,
    build_docker_image,
    is_podman,
)
from .exceptions import UserError

logger = logging.getLogger("tesseract")
docker_client = CLIDockerClient()

# Jinja2 Environment
ENV = Environment(
    loader=PackageLoader("tesseract_core.sdk", "templates"),
    undefined=StrictUndefined,
)


class LogPipe(threading.Thread):
    """Custom wrapper to support live logging from a subprocess via a pipe.

    Runs a thread that logs everything read from the pipe to the standard logger.
    Can be used as a context manager for automatic cleanup.
    """

    daemon = True

    def __init__(self, level: int) -> None:
        """Initialize the LogPipe with the given logging level."""
        super().__init__()
        self._level = level
        self._fd_read, self._fd_write = os.pipe()
        self._pipe_reader = os.fdopen(self._fd_read)
        self._captured_lines = []

    def __enter__(self) -> int:
        """Start the thread and return the write file descriptor of the pipe."""
        self.start()
        return self.fileno()

    def __exit__(self, *args: Any) -> None:
        """Close the pipe and join the thread."""
        os.close(self._fd_write)
        # Use a timeout so something weird happening in the logging thread doesn't
        # cause this to hang indefinitely
        self.join(timeout=10)
        # Do not close reader before thread is joined since there may be pending data
        # This also closes the fd_read pipe
        self._pipe_reader.close()

    def fileno(self) -> int:
        """Return the write file descriptor of the pipe."""
        return self._fd_write

    def run(self) -> None:
        """Run the thread, logging everything."""
        for line in iter(self._pipe_reader.readline, ""):
            if line.endswith("\n"):
                line = line[:-1]
            self._captured_lines.append(line)
            logger.log(self._level, line)

    @property
    def captured_lines(self) -> list[str]:
        """Return all lines captured so far."""
        return self._captured_lines


def needs_docker(func: Callable) -> Callable:
    """A decorator for functions that rely on docker daemon."""
    import functools

    @functools.wraps(func)
    def wrapper_needs_docker(*args: Any, **kwargs: Any) -> None:
        try:
            docker_client.info()
        except (APIError, RuntimeError) as ex:
            raise UserError(
                "Could not reach Docker daemon, check if it is running."
            ) from ex
        except FileNotFoundError as ex:
            raise UserError("Docker not found, check if it is installed.") from ex
        return func(*args, **kwargs)

    return wrapper_needs_docker


def get_free_port(
    within_range: tuple[int, int] = (49152, 65535),
    exclude: Sequence[int] = (),
) -> int:
    """Find a random free port to use for HTTP."""
    start, end = within_range
    if start < 0 or end > 65535 or start > end:
        raise ValueError("Invalid port range, must be between 0 and 65535")

    # Try random ports in the given range
    portlist = list(range(start, end))
    random.shuffle(portlist)
    for port in portlist:
        if port in exclude:
            continue
        # Check if the port is free
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            try:
                s.bind(("127.0.0.1", port))
            except OSError:
                # Port is already in use
                continue
            else:
                return port
    raise RuntimeError(f"No free ports found in range {start}-{end}")


def parse_requirements(
    filename: str | Path,
    session: PipSession | None = None,
    finder: PackageFinder | None = None,
    options: optparse.Values | None = None,
    constraint: bool = False,
) -> tuple[list[str], list[str]]:
    """Split local dependencies from remote ones in a pip-style requirements file.

    All CLI options that may be part of the given requiremets file are included in
    the remote dependencies.
    """
    if session is None:
        session = PipSession()

    local_dependencies = []
    remote_dependencies = []

    line_parser = get_line_parser(finder)
    parser = RequirementsFileParser(session, line_parser)

    for parsed_line in parser.parse(str(filename), constraint):
        line = linecache.getline(parsed_line.filename, parsed_line.lineno)
        line = line.strip()
        parsed_req = handle_line(
            parsed_line, options=options, finder=finder, session=session
        )
        if not hasattr(parsed_req, "requirement"):
            # this is probably a cli option like --extra-index-url, so we make
            # sure to keep it.
            remote_dependencies.append(line)
        elif parsed_line.requirement.startswith((".", "/", "file://")):
            local_dependencies.append(line)
        else:
            remote_dependencies.append(line)
    return local_dependencies, remote_dependencies


def get_runtime_dir() -> Path:
    """Get the source directory for the Tesseract runtime."""
    import tesseract_core

    return Path(tesseract_core.__file__).parent / "runtime"


def get_template_dir() -> Path:
    """Get the template directory for the Tesseract runtime."""
    import tesseract_core

    return Path(tesseract_core.__file__).parent / "sdk" / "templates"


def prepare_build_context(
    src_dir: str | Path,
    context_dir: str | Path,
    user_config: TesseractConfig,
    use_ssh_mount: bool = False,
) -> Path:
    """Populate the build context for a Tesseract.

    Generated folder structure:
    ├── Dockerfile
    ├── __tesseract_source__
    │   ├── tesseract_api.py
    │   ├── tesseract_config.yaml
    │   ├── tesseract_requirements.txt
    │   └── ... any other files in the source directory ...
    └── __tesseract_runtime__
        ├── pyproject.toml
        ├── ... any other files in the tesseract_core/runtime/meta directory ...
        └── tesseract_core
            └── runtime
                ├── __init__.py
                └── ... runtime module files ...

    Args:
        src_dir: The source directory where the Tesseract project is located.
        context_dir: The directory where the build context will be created.
        user_config: The Tesseract configuration object.
        use_ssh_mount: Whether to use SSH mount to install dependencies (prevents caching).

    Returns:
        The path to the build context directory.
    """
    context_dir = Path(context_dir)
    context_dir.mkdir(parents=True, exist_ok=True)

    copytree(src_dir, context_dir / "__tesseract_source__")

    template_name = "Dockerfile.base"
    template = ENV.get_template(template_name)

    template_values = {
        "tesseract_source_directory": "__tesseract_source__",
        "tesseract_runtime_location": "__tesseract_runtime__",
        "config": user_config,
        "use_ssh_mount": use_ssh_mount,
    }

    logger.debug(f"Generating Dockerfile from template: {template_name}")
    dockerfile_content = template.render(template_values)
    dockerfile_path = context_dir / "Dockerfile"

    logger.debug(f"Writing Dockerfile to {dockerfile_path}")

    with open(dockerfile_path, "w") as f:
        f.write(dockerfile_content)

    template_dir = get_template_dir()

    requirement_config = user_config.build_config.requirements
    copy(
        template_dir / requirement_config._build_script,
        context_dir / "__tesseract_source__" / requirement_config._build_script,
    )

    # When building from a requirements.txt we support local dependencies.
    # We separate local dep. lines from the requirements.txt and copy the
    # corresponding files into the build directory.
    local_requirements_path = context_dir / "local_requirements"
    Path.mkdir(local_requirements_path, parents=True, exist_ok=True)

    if requirement_config.provider == "python-pip":
        reqstxt = src_dir / requirement_config._filename
        if reqstxt.exists():
            local_dependencies, remote_dependencies = parse_requirements(reqstxt)
        else:
            local_dependencies, remote_dependencies = [], []

        if local_dependencies:
            for dependency in local_dependencies:
                src = src_dir / dependency
                dest = context_dir / "local_requirements" / src.name
                if src.is_file():
                    copy(src, dest)
                else:
                    copytree(src, dest)

        # We need to write a new requirements file in the build dir, where we explicitly
        # removed the local dependencies
        requirements_file_path = (
            context_dir / "__tesseract_source__" / "tesseract_requirements.txt"
        )
        with requirements_file_path.open("w", encoding="utf-8") as f:
            for dependency in remote_dependencies:
                f.write(f"{dependency}\n")

    def _ignore_pycache(_: Any, names: list[str]) -> list[str]:
        ignore = []
        if "__pycache__" in names:
            ignore.append("__pycache__")
        return ignore

    runtime_source_dir = get_runtime_dir()
    copytree(
        runtime_source_dir,
        context_dir / "__tesseract_runtime__" / "tesseract_core" / "runtime",
        ignore=_ignore_pycache,
    )
    for metafile in (runtime_source_dir / "meta").glob("*"):
        copy(metafile, context_dir / "__tesseract_runtime__")

    return context_dir


def _write_template_file(
    template_name: str,
    target_dir: Path,
    template_vars: dict,
    recipe: Path = Path("."),
    exist_ok: bool = False,
):
    """Write a template to a target directory."""
    template = ENV.get_template(str(recipe / template_name))

    target_file = target_dir / template_name

    if target_file.exists() and not exist_ok:
        raise FileExistsError(f"File {target_file} already exists")

    logger.info(f"Writing template {template_name} to {target_file}")

    with open(target_file, "w") as target_fp:
        target_fp.write(template.render(template_vars))

    return target_file


def init_api(
    target_dir: Path,
    tesseract_name: str,
    recipe: str = "base",
) -> Path:
    """Create a new empty Tesseract API module at the target location."""
    from tesseract_core import __version__ as tesseract_version

    template_vars = {
        "version": tesseract_version,
        "timestamp": datetime.datetime.now().isoformat(),
        "name": tesseract_name,
    }

    # If target dir does not exist, create it
    Path(target_dir).mkdir(parents=True, exist_ok=True)

    _write_template_file(
        "tesseract_api.py", target_dir, template_vars, recipe=Path(recipe)
    )
    _write_template_file(
        "tesseract_config.yaml", target_dir, template_vars, recipe=Path(recipe)
    )
    _write_template_file(
        "tesseract_requirements.txt", target_dir, template_vars, recipe=Path(recipe)
    )

    return target_dir / "tesseract_api.py"


def build_tesseract(
    src_dir: str | Path,
    image_tag: str | None,
    build_dir: Path | None = None,
    inject_ssh: bool = False,
    config_override: dict[tuple[str, ...], Any] | None = None,
    generate_only: bool = False,
) -> Image | Path:
    """Build a new Tesseract from a context directory.

    Args:
        src_dir: path to the Tesseract project directory, where the
          `tesseract_api.py` and `tesseract_config.yaml` files
          are located.
        image_tag: name to be used as a tag for the Tesseract image.
        build_dir: directory to be used to store the build context.
          If not provided, a temporary directory will be created.
        inject_ssh: whether or not to forward SSH agent when building the image.
        config_override: overrides for configuration options in the Tesseract.
        generate_only: only generate the build context but do not build the image.

    Returns:
        Image object representing the built Tesseract image,
        or path to build directory if `generate_only` is True.
    """
    src_dir = Path(src_dir)

    validate_tesseract_api(src_dir)
    config = get_config(src_dir)

    # Apply config overrides
    if config_override is not None:
        for path, value in config_override.items():
            c = config
            for k in path[:-1]:
                c = getattr(c, k)
            setattr(c, path[-1], value)

    image_name = config.name
    if image_tag:
        image_name += f":{image_tag}"

    source_basename = Path(src_dir).name

    if build_dir is None:
        build_dir = Path(tempfile.mkdtemp(prefix=f"tesseract_build_{source_basename}"))
        keep_build_dir = True if generate_only else False
    else:
        build_dir = Path(build_dir)
        build_dir.mkdir(exist_ok=True)
        keep_build_dir = True

    context_dir = prepare_build_context(
        src_dir, build_dir, config, use_ssh_mount=inject_ssh
    )

    if generate_only:
        logger.info(f"Build directory generated at {build_dir}, skipping build")
    else:
        logger.info("Building image ...")

    try:
        image = build_docker_image(
            path=context_dir.as_posix(),
            tag=image_name,
            dockerfile=context_dir / "Dockerfile",
            inject_ssh=inject_ssh,
            print_and_exit=generate_only,
        )
    finally:
        if not keep_build_dir:
            try:
                rmtree(build_dir)
            except OSError as exc:
                # Permission denied or already removed
                logger.info(
                    f"Could not remove temporary build directory {build_dir}: {exc}"
                )

    if generate_only:
        return build_dir

    logger.debug("Build successful")
    assert image is not None
    return image


def teardown(project_ids: Sequence[str] | None = None, tear_all: bool = False) -> None:
    """Teardown Tesseract image(s) running in a Docker Compose project or standalone containers.

    Args:
        project_ids: List of Docker Compose project IDs to teardown.
        tear_all: boolean flag to teardown all Tesseract projects.
    """
    if tear_all:
        # Identify all Tesseract projects to tear down, whether they're running in
        # Docker Compose or as standalone containers
        compose_projects = docker_client.compose.list()
        compose_containers = set()
        for project_containers in compose_projects.values():
            compose_containers.update(project_containers)
        other_containers = set(
            container.id
            for container in docker_client.containers.list()
            if container.id not in compose_containers
        )
        project_ids = [
            *compose_projects.keys(),
            *other_containers,
        ]
        if not project_ids:
            logger.info("No Tesseract projects to teardown")
            return

    if not project_ids:
        raise ValueError("project_ids must be provided if tear_all is False")

    if isinstance(project_ids, str):
        project_ids = [project_ids]

    def _is_container_id(project_id: str) -> bool:
        try:
            docker_client.containers.get(project_id)
            return True
        except ContainerError:
            return False

    for project_id in project_ids:
        if docker_client.compose.exists(project_id):
            if not docker_client.compose.down(project_id):
                raise RuntimeError(
                    f"Cannot teardown Docker Compose project with ID: {project_id}"
                )
            logger.info(
                f"Tesseracts are shutdown for Docker Compose project ID: {project_id}"
            )
        elif _is_container_id(project_id):
            container = docker_client.containers.get(project_id)
            container.remove(force=True)
            logger.info(f"Tesseract is shutdown for Docker container ID: {project_id}")
        else:
            raise ValueError(
                f"A Docker Compose project with ID {project_id} cannot be found, "
                "use `tesseract ps` to find project ID"
            )


def get_tesseract_containers() -> list[Container]:
    """Get Tesseract containers."""
    return docker_client.containers.list()


def get_tesseract_images() -> list[Image]:
    """Get Tesseract images."""
    return docker_client.images.list()


def get_project_containers(project_id: str) -> list[Container]:
    """Get containers for a given Tesseract project ID."""
    if docker_client.compose.exists(project_id):
        containers = docker_client.compose.list()[project_id]
        return [docker_client.containers.get(c) for c in containers]

    try:
        container = docker_client.containers.get(project_id)
        return [container]
    except ContainerError as exc:
        raise ValueError(
            f"A Tesseract project with ID {project_id} cannot be found, "
            "use `tesseract ps` to find project ID"
        ) from exc


def serve(
    images: list[str],
    host_ip: str = "127.0.0.1",
    ports: list[str] | None = None,
    volumes: list[str] | None = None,
    environment: dict[str, str] | None = None,
    gpus: list[str] | None = None,
    debug: bool = False,
    num_workers: int = 1,
    no_compose: bool = False,
    service_names: list[str] | None = None,
    user: str | None = None,
    input_path: str | Path | None = None,
) -> str:
    """Serve one or more Tesseract images.

    Start the Tesseracts listening on an available ports on the host.

    Args:
        images: a list of Tesseract image IDs as strings.
        host_ip: IP address to bind the Tesseracts to.
        ports: port or port range to serve each Tesseract on.
        volumes: list of paths to mount in the Tesseract container.
        environment: dictionary of environment variables to pass to the Tesseract.
        gpus: IDs of host Nvidia GPUs to make available to the Tesseracts.
        debug: Enable debug mode. This will propagate full tracebacks to the client
            and start a debugpy server in the Tesseract.
            WARNING: This may expose sensitive information, use with caution (and never in production).
        num_workers: number of workers to use for serving the Tesseracts.
        no_compose: if True, do not use Docker Compose to serve the Tesseracts.
        service_names: list of service names under which to expose each Tesseract container on the shared network.
        user: user to run the Tesseracts as, e.g. '1000' or '1000:1000' (uid:gid).
              Defaults to the current user.
        input_path: Input path to read input files from, such as local directory or S3 URI.

    Returns:
        A string representing the Tesseract project ID.
    """
    if not images or not all(isinstance(item, str) for item in images):
        raise ValueError("One or more Tesseract image IDs must be provided")

    image_ids = []
    for image_ in images:
        image = docker_client.images.get(image_)

        if not image:
            raise ValueError(f"Image ID {image_} is not a valid Docker image")
        image_ids.append(image.id)

    if ports is not None and len(ports) != len(image_ids):
        raise ValueError(
            f"Number of ports ({len(ports)}) must match number of images ({len(image_ids)})"
        )

    if service_names is not None:
        if len(service_names) != len(image_ids):
            raise ValueError(
                f"Number of service names ({len(service_names)}) must match number of images ({len(image_ids)})"
            )
        _validate_service_names(service_names)

    if user is None:
        # Use the current user if not specified
        user = f"{os.getuid()}:{os.getgid()}" if os.name != "nt" else None

    if input_path:
        if environment is None:
            environment = {}
        environment["TESSERACT_INPUT_PATH"] = "/tesseract/input_data"
        if volumes is None:
            volumes = []
        if "://" not in input_path:
            input_path = Path(input_path).resolve()
            volumes.append(f"{input_path}:/tesseract/input_data:ro")

    if no_compose:
        if len(images) > 1:
            raise ValueError(
                "Docker Compose is required to serve multiple Tesseracts. "
                f"Currently attempting to serve `{len(images)}` Tesseracts."
            )
        if service_names is not None:
            raise ValueError(
                "Tesseract service names can only be set when using Docker Compose."
            )
        args = []
        container_api_port = "8000"
        container_debugpy_port = "5678"

        args.extend(["--port", container_api_port])

        if ports:
            port = ports[0]
        else:
            port = str(get_free_port())

        if num_workers > 1:
            args.extend(["--num-workers", str(num_workers)])

        # Always bind to all interfaces inside the container
        args.extend(["--host", "0.0.0.0"])

        if host_ip == "0.0.0.0":
            ping_ip = "127.0.0.1"
        else:
            ping_ip = host_ip

        port_mappings = {f"{host_ip}:{port}": container_api_port}
        if debug:
            debugpy_port = str(get_free_port())
            port_mappings[f"{host_ip}:{debugpy_port}"] = container_debugpy_port

        logger.info(f"Serving Tesseract at http://{ping_ip}:{port}")
        logger.info(f"View Tesseract: http://{ping_ip}:{port}/docs")
        if debug:
            logger.info(f"Debugpy server listening at http://{ping_ip}:{debugpy_port}")

        parsed_volumes = _parse_volumes(volumes) if volumes else {}

        extra_args = []
        if is_podman():
            extra_args.extend(["--userns", "keep-id"])

        container = docker_client.containers.run(
            image=image_ids[0],
            command=["serve", *args],
            device_requests=gpus,
            ports=port_mappings,
            detach=True,
            volumes=parsed_volumes,
            user=user,
            environment=environment,
            extra_args=extra_args,
        )
        # wait for server to start
        timeout = 30
        while True:
            try:
                response = requests.get(f"http://{ping_ip}:{port}/health")
            except requests.exceptions.ConnectionError:
                pass
            else:
                if response.status_code == 200:
                    break

            time.sleep(0.1)
            timeout -= 0.1

            if timeout < 0:
                raise TimeoutError("Tesseract did not start in time")

        return container.name

    if is_podman() and volumes:
        raise UserError(
            "Podman does not support volume mounts in Docker Compose. "
            "Please use --no-compose / no_compose=True instead."
        )

    template = _create_docker_compose_template(
        image_ids,
        host_ip,
        service_names,
        ports,
        volumes,
        environment,
        gpus,
        num_workers,
        debug=debug,
        user=user,
    )
    compose_fname = f"docker-compose-{_id_generator()}.yml"

    with tempfile.NamedTemporaryFile(
        mode="w+",
        prefix=compose_fname,
    ) as compose_file:
        compose_file.write(template)
        compose_file.flush()

        project_name = f"tesseract-{_id_generator()}"
        try:
            docker_client.compose.up(compose_file.name, project_name)
        except APIError as exc:
            raise RuntimeError("Cannot serve Tesseracts") from exc

    return project_name


def _create_docker_compose_template(
    image_ids: list[str],
    host_ip: str = "127.0.0.1",
    service_names: list[str] | None = None,
    ports: list[str] | None = None,
    volumes: list[str] | None = None,
    environment: dict[str, str] | None = None,
    gpus: list[str] | None = None,
    num_workers: int = 1,
    debug: bool = False,
    user: str | None = None,
) -> str:
    """Create Docker Compose template."""
    services = []

    # Generate random service names for each image if not provided
    if service_names is None:
        service_names = []
        for image_id in image_ids:
            service_names.append(f"{image_id.split(':')[0]}-{_id_generator()}")

    # Get random unique ports for each image if not provided
    if ports is None:
        ports = []
        for _ in image_ids:
            taken_ports = [int(p) for p in ports if "-" not in p]
            ports.append(str(get_free_port(exclude=taken_ports)))

    # Get random unique ports for debugpy if debug mode is active
    debugpy_ports = []
    if debug:
        for _ in image_ids:
            taken_ports = [int(p) for p in ports if "-" not in p]
            debugpy_ports.append(str(get_free_port(exclude=taken_ports)))

    # Convert port ranges to fixed ports
    for i, port in enumerate(ports):
        if "-" in port:
            port_start, port_end = port.split("-")
            taken_ports = [int(p) for p in ports if "-" not in p]
            ports[i] = str(
                get_free_port(
                    within_range=(int(port_start), int(port_end)), exclude=taken_ports
                )
            )

    # Prepend host IP to ports
    ports = [f"{host_ip}:{port}" for port in ports]

    gpu_settings = None
    if gpus:
        if (len(gpus) == 1) and (gpus[0] == "all"):
            gpu_settings = "count: all"
        else:
            gpu_settings = f"device_ids: {gpus}"

    parsed_volumes = _parse_volumes(volumes) if volumes else {}

    for i, image_id in enumerate(image_ids):
        service = {
            "name": service_names[i],
            "user": user,
            "image": image_id,
            "port": f"{ports[i]}:8000",
            "volumes": parsed_volumes,
            "gpus": gpu_settings,
            "environment": {
                "TESSERACT_DEBUG": "1" if debug else "0",
                **(environment or {}),
            },
            "num_workers": num_workers,
            "debugpy_port": debugpy_ports[i] if debug else None,
        }

        services.append(service)

    external_volume_map = {}
    for source in parsed_volumes:
        # Check if the volume is a local bind-mount
        if _is_local_volume(source):
            continue

        # Check if the volume is an existing Docker volume
        try:
            docker_client.volumes.get(source)
        except NotFound:
            external_volume_map[source] = False
        else:
            # If the volume exists, it is an external Docker volume
            external_volume_map[source] = True

    template = ENV.get_template("docker-compose.yml")
    return template.render(services=services, named_volumes=external_volume_map)


def _id_generator(
    size: int = 12, chars: Sequence[str] = string.ascii_lowercase + string.digits
) -> str:
    """Generate a random ID."""
    return "".join(random.choice(chars) for _ in range(size))


def _is_local_volume(volume: str) -> bool:
    """Check if a volume is a local path."""
    return "/" in volume or "." in volume


def _parse_volumes(options: list[str]) -> dict[str, dict[str, str]]:
    """Parses volume mount strings to dict accepted by docker SDK.

    Strings of the form 'source:target:(ro|rw)' are parsed to
    `{source: {'bind': target, 'mode': '(ro|rw)'}}`.
    """

    def _parse_option(option: str):
        args = option.split(":")
        if len(args) == 2:
            source, target = args
            mode = "ro"
        elif len(args) == 3:
            source, target, mode = args
        else:
            raise ValueError(
                f"Invalid mount volume specification {option} "
                "(must be `/path/to/source:/path/totarget:(ro|rw)`)",
            )

        if _is_local_volume(source):
            if not Path(source).exists():
                raise RuntimeError(
                    f"Source path {source} does not exist, "
                    "please provide a valid local path."
                )
            # Docker doesn't like paths like ".", so we convert to absolute path here
            source = str(Path(source).resolve())
        return source, {"bind": target, "mode": mode}

    return dict(_parse_option(opt) for opt in options)


def _validate_service_names(service_names: list[str]) -> None:
    if len(set(service_names)) != len(service_names):
        raise ValueError("Service names must be unique")

    invalid_names = []
    for name in service_names:
        if not re.match(r"^[A-Za-z0-9][A-Za-z0-9-]*$", name):
            invalid_names.append(name)
        if name[-1] == "-":
            invalid_names.append(name)
    if invalid_names:
        raise ValueError(
            "Service names must contain only alphanumeric characters and hyphens, and must "
            f"not begin or end with a hyphen. Found invalid names: {invalid_names}."
        )


def run_tesseract(
    image: str,
    command: str,
    args: list[str],
    volumes: list[str] | None = None,
    gpus: list[int | str] | None = None,
    ports: dict[str, str] | None = None,
    environment: dict[str, str] | None = None,
    network: str | None = None,
    user: str | None = None,
) -> tuple[str, str]:
    """Start a Tesseract and execute a given command.

    Args:
        image: string of the Tesseract to run.
        command: Tesseract command to run, e.g. apply.
        args: arguments for the command.
        volumes: list of paths to mount in the Tesseract container.
        gpus: list of GPUs, as indices or names, to passthrough the container.
        ports: dictionary of ports to bind to the host. Key is the host port,
            value is the container port.
        environment: list of environment variables to set in the container,
            in Docker format: key=value.
        network: name of the Docker network to connect the container to.
        user: user to run the Tesseract as, e.g. '1000' or '1000:1000' (uid:gid).
            Defaults to the current user.
        input_path: Input path to read input files from, such as local directory or S3 URI.
            The input path is mounted read-only into the Tesseract at `/tesseract/input_data`.

    Returns:
        Tuple with the stdout and stderr of the Tesseract.
    """
    # Args that require rw access to the mounted volume
    output_args = {"-o", "--output-path"}
    input_args = {"-i", "--input-path"}

    cmd = [command]
    current_cmd = None

    if environment is None:
        environment = {}

    if not volumes:
        parsed_volumes = {}
    else:
        parsed_volumes = _parse_volumes(volumes)

    if user is None:
        # Use the current user if not specified
        user = f"{os.getuid()}:{os.getgid()}" if os.name != "nt" else None

    for arg in args:
        if arg.startswith("-"):
            current_cmd = arg
            cmd.append(arg)
            continue

        # Mount local output directories into Docker container as a volume
        if current_cmd in output_args and "://" not in arg:
            if arg.startswith("@"):
                raise ValueError(
                    f"Output path {arg} cannot start with '@' (used only for input files)"
                )

            local_path = Path(arg).resolve()
            local_path.mkdir(parents=True, exist_ok=True)

            if not local_path.is_dir():
                raise RuntimeError(
                    f"Path {local_path} provided as output is not a directory"
                )

            path_in_container = "/tesseract/output_data"
            arg = path_in_container

            # Bind-mount directory
            parsed_volumes[str(local_path)] = {"bind": path_in_container, "mode": "rw"}
            environment["TESSERACT_OUTPUT_PATH"] = path_in_container

        if current_cmd in input_args and "://" not in arg:
            local_path = Path(arg).resolve()
            if not local_path.is_dir():
                raise RuntimeError(
                    f"Path {local_path} provided as input is not a directory"
                )

            path_in_container = "/tesseract/input_data"
            arg = path_in_container

            # Bind-mount directory
            parsed_volumes[str(local_path)] = {"bind": path_in_container, "mode": "ro"}
            environment["TESSERACT_INPUT_PATH"] = path_in_container

        # Mount local input files marked by @ into Docker container as a volume
        elif arg.startswith("@") and "://" not in arg:
            local_path = Path(arg.lstrip("@")).resolve()

            if not local_path.is_file():
                raise RuntimeError(f"Path {local_path} provided as input is not a file")

            path_in_container = os.path.join(
                "/tesseract", f"payload{local_path.suffix}"
            )
            arg = f"@{path_in_container}"

            # Bind-mount file
            parsed_volumes[str(local_path)] = {
                "bind": path_in_container,
                "mode": "ro",
            }

        current_cmd = None
        cmd.append(arg)

    extra_args = []
    if is_podman():
        extra_args.extend(["--userns", "keep-id"])

    # Run the container
    stdout, stderr = docker_client.containers.run(
        image=image,
        command=cmd,
        volumes=parsed_volumes,
        device_requests=gpus,
        environment=environment,
        network=network,
        ports=ports,
        detach=False,
        remove=True,
        stderr=True,
        user=user,
        extra_args=extra_args,
    )
    stdout = stdout.decode("utf-8")
    stderr = stderr.decode("utf-8")
    return stdout, stderr


def logs(container_id: str) -> str:
    """Get logs from a container.

    Args:
        container_id: the ID of the container.

    Returns:
        The logs of the container.
    """
    container = docker_client.containers.get(container_id)
    return container.logs().decode("utf-8")



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from abc import ABCMeta
from enum import IntEnum
from functools import partial
from typing import (
    Annotated,
    Any,
    Optional,
    Union,
    get_args,
    get_origin,
)

import numpy as np
from pydantic import (
    AfterValidator,
    BaseModel,
    ConfigDict,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
)
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema

from tesseract_core.runtime.array_encoding import (
    AllowedDtypes,
    decode_array,
    encode_array,
    get_array_model,
    python_to_array,
)

AnnotatedType = type(Annotated[Any, Any])
EllipsisType = type(Ellipsis)


def _ensure_valid_shapedtype(expected_shape: Any, expected_dtype: Any) -> tuple:
    if not isinstance(expected_shape, (tuple, EllipsisType)):
        raise ValueError(
            "Shape in Array[<shape>, <dtype>] must be a tuple or '...' (ellipsis)"
        )

    if isinstance(expected_shape, tuple):
        for dim in expected_shape:
            if dim is not None and not isinstance(dim, int):
                raise ValueError(
                    "Shape values in Array[<shape>, <dtype>] must be integers or None"
                )

    if is_array_annotation(expected_dtype):
        expected_dtype = expected_dtype.__metadata__[0].expected_dtype

    allowed_dtypes = get_args(AllowedDtypes)

    if expected_dtype not in allowed_dtypes and expected_dtype is not None:
        raise ValueError(
            f"Invalid dtype in Array[<shape>, <dtype>]: {expected_dtype} "
            f"(must be one of {allowed_dtypes} or a scalar Array type like, Array[(), Int32])"
        )
    return expected_shape, expected_dtype


class ShapeDType(BaseModel):
    """Data structure describing an array's shape and data type."""

    shape: tuple[int, ...]
    dtype: AllowedDtypes
    # Ignore extra fields in the model, to allow encoded arrays to be passed
    model_config = ConfigDict(extra="ignore")

    def __class_getitem__(
        cls,
        key: tuple[
            Union[tuple[Optional[int], ...], EllipsisType],
            Union[AnnotatedType, str, None],
        ],
    ) -> AnnotatedType:
        expected_shape, expected_dtype = _ensure_valid_shapedtype(*key)

        def validate(shapedtype: ShapeDType) -> ShapeDType:
            """Validator to check if the shape and dtype match the expected values."""
            if isinstance(shapedtype, ShapeDType):
                shape = shapedtype.shape
                if expected_shape is Ellipsis:
                    return shapedtype

                # TODO: replace this check with `zip(... strict=True)`
                # once we stop supporting 3.9
                if len(shape) != len(expected_shape):
                    raise ValueError(
                        f"Expected shape: {expected_shape}. Found: {shape}."
                    )

                for actual, expected in zip(shape, expected_shape):
                    if expected is not None and actual != expected:
                        raise ValueError(
                            f"Expected shape: {expected_shape}. Found: {shape}."
                        )
            return shapedtype

        return Annotated[ShapeDType, AfterValidator(validate)]

    @classmethod
    def from_array_annotation(cls, obj: AnnotatedType) -> AnnotatedType:
        """Create a ShapeDType from an array annotation."""
        shape = obj.__metadata__[0].expected_shape
        dtype = obj.__metadata__[0].expected_dtype
        return cls[shape, dtype]


class ArrayFlags(IntEnum):
    """Custom flags for array annotations."""

    DIFFERENTIABLE = 1


class ArrayAnnotationType(ABCMeta):
    """Metaclass for Array type annotation to enforce repr on created types based on class variables.

    Example:
        >>> class MyArray(metaclass=ArrayAnnotationType):
        ...     expected_shape = (2, 3)
        ...     expected_dtype = "float32"

        >>> MyArray
        MyArray[(2, 3), 'float32']
    """

    def __repr__(cls) -> str:
        return f"{cls.__name__}[{cls.expected_shape!r}, {cls.expected_dtype!r}]"


def safe_issubclass(obj: Any, baseclass: type[object]) -> bool:
    """Check if obj is a subclass of baseclass in a way that never raises.

    (This is useful when obj is not guaranteed to be a type.)
    """
    try:
        return issubclass(obj, baseclass)
    except TypeError:
        return False


def _is_annotated(obj: Any) -> bool:
    """Check if an object is typing.Annotated or typing_extensions.Annotated."""
    return get_origin(obj) is Annotated


class PydanticArrayAnnotation(metaclass=ArrayAnnotationType):
    """Base class for Pydantic annotations for NumPy array types.

    This class provides Pydantic support for arrays with a fixed / polymorphic shape and dtype,
    with proper validation and serialization.

    See https://docs.pydantic.dev/latest/concepts/types/#handling-third-party-types

    When serializing or validating pydantic models that contain this annotation
    you can customize the array encoding to: plain json, base64, or binref. For
    more details see the docstring of 'Array'.
    """

    # These are class attributes that must be set when the class is created
    expected_shape: Union[tuple[int, ...], EllipsisType]
    expected_dtype: str
    flags: tuple[ArrayFlags]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        raise RuntimeError(f"{self.__class__.__name__} cannot be instantiated")

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        """This method is called by Pydantic to get the core schema for the annotated type.

        Does most of the heavy lifting for validation and serialization.
        """
        # Create a Pydantic model for the encoded array, for easier validation
        array_schema = _handler(
            get_array_model(
                cls.expected_shape,
                cls.expected_dtype,
                [flag.name for flag in cls.flags],
            )
        )

        python_to_array_ = partial(
            python_to_array,
            expected_shape=cls.expected_shape,
            expected_dtype=cls.expected_dtype,
        )
        encode_array_ = partial(
            encode_array,
            expected_shape=cls.expected_shape,
            expected_dtype=cls.expected_dtype,
        )
        decode_array_ = partial(
            decode_array,
            expected_shape=cls.expected_shape,
            expected_dtype=cls.expected_dtype,
        )

        load_from_dict_schema = core_schema.chain_schema(
            # first load / validate JSON, then decode into a NumPy array
            [
                array_schema,
                core_schema.with_info_plain_validator_function(
                    decode_array_,
                    serialization=core_schema.plain_serializer_function_ser_schema(
                        encode_array_,
                        info_arg=True,
                        return_schema=array_schema,
                    ),
                ),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=load_from_dict_schema,
            python_schema=core_schema.union_schema(
                [
                    load_from_dict_schema,
                    # when loading from Python, we also allow any array-like object
                    core_schema.no_info_plain_validator_function(python_to_array_),
                ],
                mode="left_to_right",
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                encode_array_,
                info_arg=True,
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        """This method is called by Pydantic to get the JSON schema for the annotated type."""
        return handler(_core_schema)


class Array:
    """Generic Pydantic type annotation for a multi-dimensional array with a fixed shape and dtype.

    Arrays will be broadcasted to the expected shape and dtype during validation,
    but dimensions must match exactly.

    Polymorphic dimensions are supported by using `None` in the shape tuple.
    To indicate a scalar, use an empty tuple.

    Arrays of any shape and rank can be represented by using `...` (ellipsis) as the shape.

    Example:
        >>> class MyModel(BaseModel):
        ...     int_array: Array[(2, 3), Int32]
        ...     float_array: Array[(None, 3), Float64]
        ...     scalar_int: Array[(), Int16]
        ...     any_shape_array: Array[..., Float32]


    You can serialize to (and validate from) different array encodings.

        >>> model = MyModel(
        ...     int_array=np.array([[1, 2, 3], [4, 5, 6]]),
        ...     float_array=np.array([[1.0, 2.0, 3.0]]),
        ...     scalar_int=np.int32(42),
        ...     any_shape_array=np.array([True, False, True]).reshape(1, 1, 3),
        ... )

        >>> model.model_dump_json(context={"array_encoding": "json"})

        >>> model.model_dump_json(context={"array_encoding": "base64"})

    or to binref:

        >>> model.model_dump_json(
        ...     context={
        ...         "array_encoding": "binref",
        ...         "base_dir": "path/to/base",
        ...         "max_file_size": 10**8,
        ...     }
        ... )

    In the 'binref' case you have to provide a base_dir to save/load binary
    (.bin) files. The .bin file(s) are written to `context['base_dir'] /
    f"{context['__binref_uuid']}.bin"`.  The '__binref_uuid' is considered an internal
    variable and should not be modified manually!  You can set a 'max_file_size'
    for the binary files.  When this file size (in bytes) is reached, a new
    __binref_uuid (i.e. a new .bin) is created to append array data to.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        clsname = self.__class__.__name__
        raise RuntimeError(
            f"{clsname} cannot be instantiated directly, perhaps you meant to use `{clsname}[(shape), dtype]`?"
        )

    def __class_getitem__(
        cls,
        key: tuple[
            Union[tuple[Optional[int], ...], EllipsisType],
            Union[AnnotatedType, str, None],
        ],
    ) -> AnnotatedType:
        """Create a new type annotation based on the given shape and dtype."""
        expected_shape, expected_dtype = _ensure_valid_shapedtype(*key)
        classvars = {
            "expected_shape": expected_shape,
            "expected_dtype": expected_dtype,
            "flags": (),
            "__module__": cls.__module__,
        }
        model = type(cls.__name__, (PydanticArrayAnnotation,), classvars)
        return Annotated[np.ndarray, model]


def is_array_annotation(obj: Any) -> bool:
    """Check if an object is a Pydantic array type annotation."""
    if _is_annotated(obj):
        metadata = obj.__metadata__[0]
        if safe_issubclass(metadata, PydanticArrayAnnotation):
            return True
    return False


class Differentiable:
    """Type annotation for a differentiable array.

    Example:
        >>> class MyModel(BaseModel):
        ...     array: Differentiable[Array[(None, 3), Float64]]
    """

    def __class_getitem__(cls, key: Any) -> AnnotatedType:
        """Mark wrapped array type as differentiable."""
        if not is_array_annotation(key):
            raise ValueError("Differentiable can only be applied to Array types")

        arr = key.__metadata__[0]
        # Create a new array type with the DIFFERENTIABLE flag, to not modify the original type in-place
        newarr = type(
            arr.__name__, (arr,), {"flags": (*arr.flags, ArrayFlags.DIFFERENTIABLE)}
        )
        return Annotated[key.__origin__, newarr]


def is_differentiable(obj: Any) -> bool:
    """Check if an object is a Differentiable array type annotation."""
    if is_array_annotation(obj):
        return ArrayFlags.DIFFERENTIABLE in obj.__metadata__[0].flags

    return False


# Export concrete scalar types
Float16 = Array[(), "float16"]
Float32 = Array[(), "float32"]
Float64 = Array[(), "float64"]
Int8 = Array[(), "int8"]
Int16 = Array[(), "int16"]
Int32 = Array[(), "int32"]
Int64 = Array[(), "int64"]
Bool = Array[(), "bool"]
UInt8 = Array[(), "uint8"]
UInt16 = Array[(), "uint16"]
UInt32 = Array[(), "uint32"]
UInt64 = Array[(), "uint64"]
Complex64 = Array[(), "complex64"]
Complex128 = Array[(), "complex128"]



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import os
import random
import time
from pathlib import Path

import pytest
import yaml
from jinja2.exceptions import TemplateNotFound
from typeguard import suppress_type_checks

from tesseract_core.sdk import engine
from tesseract_core.sdk.api_parse import (
    TesseractConfig,
    validate_tesseract_api,
)
from tesseract_core.sdk.cli import AVAILABLE_RECIPES
from tesseract_core.sdk.docker_client import Image, NotFound
from tesseract_core.sdk.exceptions import UserError


def test_prepare_build_context(tmp_path_factory):
    """Test we can create a dockerfile."""
    src_dir = tmp_path_factory.mktemp("src")
    (src_dir / "foo").touch()
    build_dir = tmp_path_factory.mktemp("build")
    default_config = TesseractConfig(name="foobar")
    engine.prepare_build_context(src_dir, build_dir, default_config)
    assert (build_dir / "__tesseract_source__" / "foo").exists()
    assert (build_dir / "__tesseract_runtime__").exists()
    assert (build_dir / "Dockerfile").exists()


@pytest.mark.parametrize("generate_only", [True, False])
def test_build_tesseract(dummy_tesseract_package, mocked_docker, generate_only, caplog):
    """Test we can build an image for a package and keep build directory."""
    src_dir = dummy_tesseract_package
    image_name = "unit_vectoradd"
    image_tag = "42"

    with caplog.at_level(logging.INFO):
        out = engine.build_tesseract(
            src_dir,
            image_tag,
            generate_only=generate_only,
        )

    if generate_only:
        assert isinstance(out, Path)

        # Check if Dockerfile is there
        dockerfile_path = out / "Dockerfile"
        assert dockerfile_path.exists()

        # Check stdout if it contains the correct docker build command
        assert "docker buildx build" in caplog.text
        assert str(out) in caplog.text
    else:
        assert isinstance(out, Image)
        assert out.attrs == mocked_docker.images.get(image_name).attrs


@pytest.mark.parametrize("recipe", [None, *AVAILABLE_RECIPES])
def test_init(tmpdir, recipe):
    """Test the initialization of a tesseract from the template."""
    if recipe:
        api_path = engine.init_api(
            target_dir=Path(tmpdir) / "test_dir", tesseract_name="foo", recipe=recipe
        )
    else:
        api_path = engine.init_api(
            target_dir=Path(tmpdir) / "test_dir", tesseract_name="foo"
        )

    # Make sure that all tesseract related files are created
    assert api_path.exists()
    assert (tmpdir / "test_dir/tesseract_requirements.txt").exists()
    assert (tmpdir / "test_dir/tesseract_config.yaml").exists()

    # Ensure the name in the config is correct
    with open(tmpdir / "test_dir/tesseract_config.yaml") as config_yaml:
        test = yaml.safe_load(config_yaml)
        assert test["name"] == "foo"

    # Ensure template passes validation
    validate_tesseract_api(api_path.parent)

    if not recipe:
        # Ensure it still passes when commenting in optional endpoints
        with open(api_path) as f:
            api_code = f.read()

        start_idx = api_code.find("# Optional endpoints")
        assert start_idx != -1

        api_code = api_code[: start_idx + 1] + api_code[start_idx + 1 :].replace(
            "# ", ""
        )

        with open(api_path, "w") as f:
            f.write(api_code)

    validate_tesseract_api(api_path.parent)


def test_init_bad_recipe(tmpdir):
    """Test the initialization of a tesseract with a bad recipe.

    This does not check for pretty terminal output or typer validation.
    But ensures that an error is raised if there are missing template files.
    """
    with pytest.raises(TemplateNotFound):
        engine.init_api(
            target_dir=Path(tmpdir) / "test_dir",
            tesseract_name="foo",
            recipe="recipewillneverexist",
        )


def test_run_tesseract(mocked_docker):
    """Test running a tesseract."""
    res_out, res_err = engine.run_tesseract(
        "foobar", "apply", ['{"inputs": {"a": [1, 2, 3], "b": [4, 5, 6]}}']
    )

    # Mocked docker just returns the kwargs to `docker run` as json
    res = json.loads(res_out)
    assert res["command"] == ["apply", '{"inputs": {"a": [1, 2, 3], "b": [4, 5, 6]}}']
    assert res["image"] == "foobar"

    # Also check that stderr is captured
    assert res_err == "hello tesseract"

    # Check that we did not request GPUs by accident
    assert res["device_requests"] is None


def test_run_gpu(mocked_docker):
    """Test running a tesseract with all available GPUs."""
    res_out, _ = engine.run_tesseract(
        "foobar",
        "apply",
        ['{"inputs": {"a": [1, 2, 3], "b": [4, 5, 6]}}'],
        gpus=["all"],
    )

    res = json.loads(res_out)
    assert res["device_requests"] == ["all"]


def test_run_tesseract_file_input(mocked_docker, tmpdir):
    """Test running a tesseract with file input / output."""
    outdir = Path(tmpdir) / "output"
    outdir.mkdir()

    infile = Path(tmpdir) / "input.json"
    infile.touch()

    res, _ = engine.run_tesseract(
        "foobar",
        "apply",
        [f"@{infile}", "--output-path", str(outdir)],
    )

    # Mocked docker just returns the kwargs to `docker run` as json
    res = json.loads(res)
    assert res["command"] == [
        "apply",
        "@/tesseract/payload.json",
        "--output-path",
        "/tesseract/output_data",
    ]
    assert res["image"] == "foobar"
    assert res["volumes"].keys() == {str(infile), str(outdir)}

    # Test the same with a folder mount
    res, _ = engine.run_tesseract(
        "foobar",
        "apply",
        [f"@{infile}", "--output-path", str(outdir)],
        volumes=[f"{tmpdir}:/path/in/container:ro"],
    )
    res = json.loads(res)
    assert res["volumes"].keys() == {str(infile), str(outdir), f"{tmpdir}"}
    assert res["volumes"][f"{tmpdir}"] == {
        "mode": "ro",
        "bind": "/path/in/container",
    }

    # Test the same but with --input_path
    indir = tmpdir / "input_path"
    indir.mkdir()
    res, _ = engine.run_tesseract(
        "foobar",
        "apply",
        [f"@{infile}", "--output-path", str(outdir), "--input-path", str(indir)],
    )
    res = json.loads(res)
    assert res["volumes"].keys() == {str(outdir), str(indir), str(infile)}
    assert res["volumes"][str(indir)] == {
        "mode": "ro",
        "bind": "/tesseract/input_data",
    }


def test_serve_tesseracts_invalid_input_args(mocked_docker):
    """Test input validation logic for multi-tesseract serve."""
    with suppress_type_checks():
        with pytest.raises(ValueError):
            engine.serve([None])

        with pytest.raises(ValueError):
            engine.serve([])

        with pytest.raises(ValueError):
            engine.serve([None, "vectoradd"])

        with pytest.raises(ValueError):
            engine.teardown(None)

        with pytest.raises(ValueError):
            engine.serve(["vectoradd"], ports=["8080", "8081"])

        with pytest.raises(ValueError):
            engine.serve(["vectoradd"], service_names=["A", "B"])

        with pytest.raises(ValueError):
            engine.serve(
                ["vectoradd", "vectoradd"],
                no_compose=True,
                service_names=["VA1", "VA2"],
            )

        with pytest.raises(ValueError):
            engine.serve(["vectoradd", "vectoradd"], service_names=["dupe", "dupe"])

        with pytest.raises(ValueError):
            engine.serve(["vectoradd"], service_names=["inval$id-domain-name"])

        with pytest.raises(ValueError):
            engine.serve(["vectoradd"], service_names=["-invalid-name"])


def test_get_tesseract_images(mocked_docker):
    tesseract_images = engine.get_tesseract_images()
    assert len(tesseract_images) == 2


def test_get_tesseract_containers(mocked_docker):
    tesseract_containers = engine.get_tesseract_containers()
    assert len(tesseract_containers) == 1


def test_serve_tesseracts(mocked_docker):
    """Test multi-tesseract serve."""
    # Serve valid
    project_name_single_tesseract = engine.serve(["vectoradd"])
    assert project_name_single_tesseract

    # Teardown valid
    engine.teardown(project_name_single_tesseract)

    # Multi-serve valid
    project_name_multi_tesseract = engine.serve(["vectoradd", "vectoradd"])
    assert project_name_multi_tesseract

    # Multi-teardown valid
    engine.teardown(project_name_multi_tesseract)

    # Tear down invalid
    with pytest.raises(NotFound):
        engine.teardown("invalid_project_name")

    # Serve with gpus
    project_name_multi_tesseract = engine.serve(["vectoradd"], gpus=["1", "3"])
    assert project_name_multi_tesseract

    # Serve and specify tesseract service names
    project_name_multi_tesseract = engine.serve(
        ["vectoradd", "vectoradd"], service_names=["VA1", "VA2"]
    )
    assert project_name_multi_tesseract


@pytest.mark.parametrize("no_compose", [True, False])
def test_serve_tesseract_volumes(mocked_docker, tmpdir, no_compose):
    """Test running a tesseract with volumes."""
    # Test with a single volume
    res = engine.serve(
        ["foobar"],
        volumes=[f"{tmpdir}:/path/in/container:ro"],
        no_compose=no_compose,
    )

    if no_compose:
        # Currently no good way to test return value of serve with no_compose=True
        # since it returns a project ID.
        res = json.loads(res)
        assert res["volumes"].keys() == {f"{tmpdir}"}
        assert res["volumes"][f"{tmpdir}"] == {
            "mode": "ro",
            "bind": "/path/in/container",
        }

    # Test with a named volume
    res = engine.serve(
        ["foobar"],
        volumes=["my_named_volume:/path/in/container:ro"],
        no_compose=no_compose,
    )

    if no_compose:
        res = json.loads(res)
        assert res["volumes"].keys() == {"my_named_volume"}
        assert res["volumes"]["my_named_volume"] == {
            "mode": "ro",
            "bind": "/path/in/container",
        }

    with pytest.raises(RuntimeError):
        # Test with a volume that does not exist
        engine.serve(
            ["foobar"],
            volumes=["/non/existent/path:/path/in/container:ro"],
            no_compose=no_compose,
        )


def test_needs_docker(mocked_docker, monkeypatch):
    @engine.needs_docker
    def run_something_with_docker():
        pass

    # Happy case
    run_something_with_docker()

    # Sad case
    def raise_docker_error(*args, **kwargs):
        raise RuntimeError("No Docker")

    monkeypatch.setattr(mocked_docker, "info", raise_docker_error)

    with pytest.raises(UserError):
        run_something_with_docker()


def test_logpipe(caplog):
    # Verify that logging in a separate thread works as intended
    from tesseract_core.sdk.engine import LogPipe

    caplog.set_level(logging.INFO, logger="tesseract")

    logged_lines = []
    for _ in range(100):
        msg_length = 2 ** random.randint(1, 12)
        msg = "".join(random.choices("abcdefghijklmnopqrstuvwxyz", k=msg_length))
        logged_lines.append(msg)

    logpipe = LogPipe(logging.INFO)
    with logpipe:
        fd = os.fdopen(logpipe.fileno(), "w", closefd=False)
        for line in logged_lines:
            print(line, file=fd)
            time.sleep(random.random() / 100)
        fd.flush()

    assert logpipe.captured_lines == logged_lines
    assert caplog.record_tuples == [
        ("tesseract", logging.INFO, line) for line in logged_lines
    ]


def test_parse_requirements(tmpdir):
    reqs = """
    --extra-index-url https://download.pytorch.org/whl/cpu
    torch==2.5.1

    --find-links https://data.pyg.org/whl/torch-2.5.1+cpu.html
    torch_scatter==2.1.2+pt25cpu

    ./internal_packages/foobar
    """
    reqs_file = Path(tmpdir) / "requirements.txt"
    with open(reqs_file, "w") as fi:
        fi.write(reqs)
    locals, remotes = engine.parse_requirements(reqs_file)

    assert locals == [
        "./internal_packages/foobar",
    ]
    assert remotes == [
        "--extra-index-url https://download.pytorch.org/whl/cpu",
        "torch==2.5.1",
        "--find-links https://data.pyg.org/whl/torch-2.5.1+cpu.html",
        "torch_scatter==2.1.2+pt25cpu",
    ]



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Sequence

import numpy as np
from pydantic import BaseModel, Field

from tesseract_core.runtime import Array, Differentiable, Float32
from tesseract_core.runtime.experimental import LazySequence


# input-schema-label-begin
class InputSchema(BaseModel):
    # NOTE: no file references here
    data: LazySequence[Differentiable[Array[(None, 3), Float32]]] = Field(
        description="Data to be processed."
    )


# input-schema-label-end


class OutputSchema(BaseModel):
    data: Sequence[Array[(None, 3), Float32]] = Field(description="Processed data.")
    data_sum: Differentiable[Array[(3,), Float32]] = Field(
        description="Sum of all data samples."
    )


def apply(inputs: InputSchema) -> OutputSchema:
    """Process data samples and compute their sum."""
    out_data = []
    data_sum = np.zeros(3)

    # iterating over inputs.data loads its contents one by one
    for data in inputs.data:
        # we only keep processed data here for demonstration
        out_data.append(data * 2)
        data_sum += data.sum(axis=0)

    return OutputSchema(data=out_data, data_sum=data_sum)


def jacobian(
    inputs: InputSchema,
    jac_inputs: set[str],
    jac_outputs: set[str],
):
    assert jac_outputs == {"data_sum"}

    # jac_inputs is a set of keys of the form "data.[i]".
    out = {"data_sum": {}}
    for key in jac_inputs:
        key_parts = key.split(".")
        assert key_parts[0] == "data"
        assert len(key_parts) == 2

        idx = int(key_parts[1][1:-1])
        jac = np.zeros((3, *inputs.data[idx].shape))
        jac[0, :, 0] = 1.0
        jac[1, :, 1] = 1.0
        jac[2, :, 2] = 1.0
        out["data_sum"][key] = jac

    return out



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import re
from collections.abc import Callable, Mapping, Sequence
from copy import deepcopy
from typing import Any, Optional, Union

from pydantic import BaseModel


def path_to_index_op(path: str) -> tuple[str, Union[int, str]]:
    """Converts a path string to a tuple of operation and index."""
    seq_idx_re = re.match(r"\[(\d+)\]", path)
    if seq_idx_re:
        return ("seq", int(seq_idx_re.group(1)))

    dict_idx_re = re.match(r"\{(.+)\}", path)
    if dict_idx_re:
        return ("dict", dict_idx_re.group(1))

    getattr_re = re.match(r"(\w+)", path)
    if getattr_re:
        return ("getattr", getattr_re.group(1))

    raise ValueError(f"Invalid path: {path}")


def get_at_path(tree: Any, path: str) -> Any:
    """Gets the value at a path in a nested pytree.

    Paths can have a structure like `a.b.[0].c.{key}` where:
    - `a` is an attribute / key of the input tree
    - `b.[0]` is the first element of the list `b`
    - `c.{key}` is the value of the key `key` in the dictionary `c`
    """
    split_path = path.split(".")

    def _get_recursive(tree: Any, path: list[str]) -> Any:
        if not path:
            return tree

        key, path = path[0], path[1:]
        method, idx = path_to_index_op(key)
        if method in ("seq", "dict"):
            return _get_recursive(tree[idx], path)
        elif method == "getattr":
            if hasattr(tree, key):
                return _get_recursive(getattr(tree, key), path)
            elif isinstance(tree, Mapping):
                # If the key is not an attribute, try to access it as a key in a dictionary
                # This is useful for accessing keys of models that have been dumped to dictionaries
                return _get_recursive(tree[key], path)
            else:
                raise AttributeError(f"Attribute {key} not found in {tree}")
        else:
            raise AssertionError(f"Invalid method: {method}")

    return _get_recursive(tree, split_path)


def set_at_path(tree: Any, values: dict[str, Any]) -> Any:
    """Sets the value at a collection of paths in a nested pytree.

    `values` argument is a flat dictionary with paths as keys and values as values.

    Paths can have a structure like `a.b.[0].c.{key}` where:
    - `a` is an attribute / key of the input tree
    - `b.[0]` is the first element of the list `b`
    - `c.{key}` is the value of the key `key` in the dictionary `c`
    """
    tree = deepcopy(tree)

    def _set_recursive(tree: Any, path: list[str], value: Any) -> Any:
        key, path = path[0], path[1:]
        method, idx = path_to_index_op(key)
        if method in ("seq", "dict"):
            if not path:
                tree[idx] = value
                return
            return _set_recursive(tree[idx], path, value)
        elif method == "getattr":
            if hasattr(tree, key):
                if not path:
                    setattr(tree, key, value)
                    return
                return _set_recursive(getattr(tree, key), path, value)
            elif isinstance(tree, Mapping):
                # If the key is not an attribute, try to access it as a key in a dictionary
                # This is useful for accessing keys of models that have been dumped to dictionaries
                if not path:
                    tree[key] = value
                    return
                return _set_recursive(tree[key], path, value)
            else:
                raise AttributeError(f"Attribute {key} not found in {tree}")
        else:
            raise AssertionError(f"Invalid method: {method}")

    for path, value in values.items():
        split_path = path.split(".")
        _set_recursive(tree, split_path, value)

    return tree


def flatten_with_paths(
    tree: Union[Mapping, Sequence, BaseModel],
    include_paths: set[str],
) -> dict[str, Any]:
    """Filter and flatten a nested PyTree by extracting only the specified paths.

    Returns a dictionary with the specified keys and corresponding values.
    """
    out = {}
    for path in include_paths:
        out[path] = get_at_path(tree, path)
    return out


def filter_func(
    func: Callable[[dict], dict],
    default_inputs: dict,
    output_paths: Optional[set[str]] = None,
    input_paths: Optional[set[str]] = None,
) -> Callable:
    """Modifies a function that operates on pytrees to operate on flat {path: value} or positional args instead.

    The returned function accepts either a dictionary `{input_path: value}` if `input_paths` is None or
    positional arguments in the same order as input_paths.
    The function will update the default inputs with the new values.
    It will then call the original function with the updated inputs and return a dictionary
    `{output_path: value}` if output_paths is not None, or the full unmodified output otherwise.

    Args:
        func: The original function that accepts a single pytree as input and returns a single pytree as output.
        default_inputs: The default input pytree to the function. Also used to determine the structure of the inputs.
        output_paths: The subset of paths of the outputs that the modified function returns.
            If None, the full output is returned unmodified.
        input_paths: The paths that positional arguments correspond to.
            If None, a single dictionary argument is expected by the modified function.
    """

    def filtered_func(*args: Any) -> dict:
        if input_paths:
            # TODO: replace this check with `zip(... strict=True)`
            # once we stop supporting 3.9
            assert len(input_paths) == len(args), (
                f"Mismatch between number of given paths {len(input_paths)} and args {len(args)}."
            )
            new_inputs = dict(zip(input_paths, args))
        else:
            if len(args) != 1:
                raise ValueError("Expected a single dictionary argument")
            new_inputs = args[0]

        updated_inputs = set_at_path(default_inputs, new_inputs)
        outputs = func(updated_inputs)
        if output_paths:
            outputs = flatten_with_paths(outputs, output_paths)
        return outputs

    return filtered_func
